from collections import namedtuple
import traceback
import logging
from enum import Enum

import attr

from .. import entities, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class AppScope(str, Enum):
    """ The scope of the app.

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - SYSTEM
         - Dataloop internal app
       * - PROJECT
         - Project app
    """
    SYSTEM = 'system'
    PROJECT = 'project'


@attr.s
class App(entities.BaseEntity):
    id = attr.ib(type=str)
    name = attr.ib(type=str)
    url = attr.ib(type=str)
    created_at = attr.ib(type=str)
    updated_at = attr.ib(type=str)
    creator = attr.ib(type=str)
    project_id = attr.ib(type=str)
    org_id = attr.ib(type=str)
    dpk_name = attr.ib(type=str)
    dpk_version = attr.ib(type=str)
    composition_id = attr.ib(type=str)
    scope = attr.ib(type=str)
    routes = attr.ib(type=dict)
    custom_installation = attr.ib(type=dict)
    metadata = attr.ib(type=dict)
    status = attr.ib(type=entities.CompositionStatus)
    settings = attr.ib(type=dict)

    # sdk
    _project = attr.ib(type=entities.Project, repr=False)
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)
    integrations = attr.ib(type=list, default=None)

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories', field_names=['projects', 'apps', 'compositions'])
        return reps(
            projects=repositories.Projects(client_api=self._client_api),
            apps=repositories.Apps(client_api=self._client_api, project=self._project, project_id=self.project_id),
            compositions=repositories.Compositions(client_api=self._client_api, project=self._project)
        )

    @property
    def project(self):
        if self._project is None:
            self._project = self.projects.get(project_id=self.project_id)
        return self._project

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def apps(self):
        assert isinstance(self._repositories.apps, repositories.Apps)
        return self._repositories.apps

    @property
    def compositions(self):
        assert isinstance(self._repositories.compositions, repositories.Compositions)
        return self._repositories.compositions

    def uninstall(self):
        """
        Uninstall an app installed app from the project.

        **Example**
        .. code-block:: python
            succeed = app.uninstall()
        """
        return self.apps.uninstall(app=self)

    def update(self):
        """
        Update the current app to the new configuration

        :return bool whether the operation ran successfully or not

        **Example**
        .. code-block:: python
            succeed = app.update()
        """
        return self.apps.update(self)

    def resume(self):
        """
        Resume the current app

        :return bool whether the operation ran successfully or not

        **Example**
        .. code-block:: python
            succeed = app.resume()
        """
        return self.apps.resume(self)

    def pause(self):
        """
        Pause the current app

        :return bool whether the operation ran successfully or not

        **Example**
        .. code-block:: python
            succeed = app.pause()
        """
        return self.apps.pause(self)

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json:  platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            package = App.from_json(_json=_json,
                                    client_api=client_api,
                                    project=project,
                                    is_fetched=is_fetched)
            status = True
        except Exception:
            package = traceback.format_exc()
            status = False
        return status, package

    def to_json(self):
        _json = {}
        if self.id is not None:
            _json['id'] = self.id
        if self.name is not None:
            _json['name'] = self.name
        if self.url is not None:
            _json['url'] = self.url
        if self.created_at is not None:
            _json['createdAt'] = self.created_at
        if self.updated_at is not None:
            _json['updatedAt'] = self.updated_at
        if self.creator is not None:
            _json['creator'] = self.creator
        if self.project_id is not None:
            _json['projectId'] = self.project_id
        if self.org_id is not None:
            _json['orgId'] = self.org_id
        if self.dpk_name is not None:
            _json['dpkName'] = self.dpk_name
        if self.dpk_version is not None:
            _json['dpkVersion'] = self.dpk_version
        if self.composition_id is not None:
            _json['compositionId'] = self.composition_id
        if self.scope is not None:
            _json['scope'] = self.scope
        if self.routes != {}:
            _json['routes'] = self.routes
        if self.custom_installation != {}:
            _json['customInstallation'] = self.custom_installation
        if self.metadata is not None:
            _json['metadata'] = self.metadata
        if self.status is not None:
            _json['status'] = self.status
        if self.settings != {}:
            _json['settings'] = self.settings
        if self.integrations is not None:
            _json['integrations'] = self.integrations

        return _json

    @classmethod
    def from_json(cls, _json, client_api: ApiClient, project: entities.Project=None, is_fetched=True):
        app = cls(
            id=_json.get('id', None),
            name=_json.get('name', None),
            url=_json.get('url', None),
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            creator=_json.get('creator', None),
            project_id=_json.get('projectId', None),
            org_id=_json.get('orgId', None),
            dpk_name=_json.get('dpkName', None),
            dpk_version=_json.get('dpkVersion', None),
            composition_id=_json.get('compositionId', None),
            scope=_json.get('scope', None),
            routes=_json.get('routes', {}),
            custom_installation=_json.get('customInstallation', {}),
            client_api=client_api,
            project=project,
            metadata=_json.get('metadata', None),
            status=_json.get('status', None),
            settings=_json.get('settings', {}),
            integrations=_json.get('integrations', None)
        )
        app.is_fetched = is_fetched
        return app



import numpy as np
import urllib.parse
import logging
import json
import os
import io
import copy
from typing import Generator, Tuple, Optional
from collections import deque
from concurrent.futures import ThreadPoolExecutor, wait, FIRST_COMPLETED
from bson import ObjectId


from enum import Enum
from .. import exceptions, entities

logger = logging.getLogger(name="dtlpy")


class FiltersKnownFields(str, Enum):
    DIR = "dir"
    ANNOTATED = "annotated"
    FILENAME = "filename"
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"
    LABEL = "label"
    NAME = "name"
    HIDDEN = "hidden"
    TYPE = "type"


class FiltersResource(str, Enum):
    ITEM = "items"
    ANNOTATION = "annotations"
    EXECUTION = "executions"
    PACKAGE = "packages"
    DPK = "dpks"
    APP = "apps"
    SERVICE = "services"
    TRIGGER = "triggers"
    MODEL = "models"
    WEBHOOK = "webhooks"
    RECIPE = "recipe"
    DATASET = "datasets"
    ONTOLOGY = "ontology"
    TASK = "tasks"
    PIPELINE = "pipeline"
    PIPELINE_EXECUTION = "pipelineState"
    COMPOSITION = "composition"
    FEATURE = "feature_vectors"
    FEATURE_SET = "feature_sets"
    ORGANIZATIONS = "organizations"
    DRIVERS = "drivers"
    SETTINGS = "setting"
    RESOURCE_EXECUTION = "resourceExecution"
    METRICS = ("metrics",)
    SERVICE_DRIVER = ("serviceDrivers",)
    COMPUTE = "compute"


class FiltersOperations(str, Enum):
    OR = "or"
    AND = "and"
    IN = "in"
    NOT_EQUAL = "ne"
    EQUAL = "eq"
    GREATER_THAN = "gt"
    LESS_THAN = "lt"
    EXISTS = "exists"
    MATCH = "match"
    NIN = "nin"
    GREATER_THAN_OR_EQUAL = "gte"
    LESS_THAN_OR_EQUAL = "lte"


class FiltersMethod(str, Enum):
    OR = "or"
    AND = "and"


class FiltersOrderByDirection(str, Enum):
    DESCENDING = "descending"
    ASCENDING = "ascending"


class Filters:
    """
    Filters entity to filter items from pages in platform
    """

    def __init__(
        self,
        field=None,
        values=None,
        operator: FiltersOperations = None,
        method: FiltersMethod = None,
        custom_filter=None,
        resource: FiltersResource = FiltersResource.ITEM,
        use_defaults=True,
        context=None,
        page_size=None,
    ):
        if page_size is None:
            if resource in [FiltersResource.EXECUTION, FiltersResource.PIPELINE_EXECUTION, FiltersResource.DPK]:
                page_size = 100
            else:
                page_size = 1000

        self.or_filter_list = list()
        self.and_filter_list = list()
        self._unique_fields = list()
        self.custom_filter = custom_filter
        self.known_operators = ["or", "and", "in", "ne", "eq", "gt", "lt", "exists"]
        self._resource = resource
        self.page = 0
        self.page_size = page_size
        self.method = FiltersMethod.AND
        self.sort = dict()
        self.join = None
        self.recursive = True

        # system only - task and assignment attributes
        self._user_query = "true"
        self._ref_task = False
        self._ref_assignment = False
        self._ref_op = None
        self._ref_assignment_id = None
        self._ref_task_id = None
        self._system_space = None

        self._use_defaults = use_defaults
        self.__add_defaults()
        self.context = context

        if field is not None:
            self.add(field=field, values=values, operator=operator, method=method)

    def __validate_page_size(self):
        max_page_size = self.__max_page_size
        if self.page_size > max_page_size:
            logger.warning(
                "Cannot list {} with page size greater than {}. Changing page_size to {}.".format(
                    self.resource, max_page_size, max_page_size
                )
            )
            self.page_size = max_page_size

    @property
    def __max_page_size(self):
        page_size = 1000
        if self.resource in [FiltersResource.EXECUTION, FiltersResource.PIPELINE_EXECUTION]:
            page_size = 100
        return page_size

    @property
    def resource(self):
        return f"{self._resource.value}" if isinstance(self._resource, FiltersResource) else f"{self._resource}"

    @resource.setter
    def resource(self, resource):
        self._resource = resource
        self.reset()
        self.__add_defaults()

    @property
    def system_space(self):
        return self._system_space

    @system_space.setter
    def system_space(self, val: bool):
        self._system_space = val

    def reset(self):
        self.or_filter_list = list()
        self.and_filter_list = list()
        self._unique_fields = list()
        self.custom_filter = None
        self.page = 0
        self.page_size = 1000
        self.method = FiltersMethod.AND
        self.sort = dict()
        self.join = None
        self.recursive = True
        self._nullify_refs()

    def _nullify_refs(self):
        self._ref_task = False
        self._ref_assignment = False
        self._ref_op = None
        self._ref_assignment_id = None
        self._ref_task_id = None

    def add(self, field, values, operator: FiltersOperations = None, method: FiltersMethod = None):
        """
        Add filter

        :param str field: Metadata field / attribute
        :param values: field values
        :param dl.FiltersOperations operator: optional - in, gt, lt, eq, ne
        :param dl.FiltersMethod method: Optional - or/and

        **Example**:

        .. code-block:: python

            filter.add(field='metadata.user', values=['1','2'], operator=dl.FiltersOperations.IN)
        """
        if method is None:
            method = self.method
        if "metadata.system.refs.metadata" in field and self.resource == FiltersResource.ITEM:
            logger.warning(
                "Filtering by metadata.system.refs.metadata may cause incorrect results. please use match operator"
            )

        # create SingleFilter object and add to self.filter_list
        if method == FiltersMethod.OR:
            self.or_filter_list.append(SingleFilter(field=field, values=values, operator=operator))
        elif method == FiltersMethod.AND:
            self.__override(field=field, values=values, operator=operator)
        else:
            raise exceptions.PlatformException(
                error="400", message="Unknown method {}, please select from: or/and".format(method)
            )

    def __override(self, field, values, operator=None):
        if field in self._unique_fields:
            indices_to_remove = []
            for i_single_filter, single_filter in enumerate(self.and_filter_list):
                if single_filter.field == field:
                    indices_to_remove.append(i_single_filter)
            
            # Remove indices in descending order to avoid IndexError
            # When removing items, indices shift down, so we must remove from highest to lowest
            for index in sorted(indices_to_remove, reverse=True):
                self.and_filter_list.pop(index)
        self.and_filter_list.append(SingleFilter(field=field, values=values, operator=operator))

    def generate_url_query_params(self, url):
        """
        generate url query params

        :param str url:
        """
        url = "{}?".format(url)
        for f in self.and_filter_list:
            if isinstance(f.values, list):
                url = "{}{}={}&".format(url, f.field, ",".join(f.values))
            else:
                url = "{}{}={}&".format(url, f.field, f.values)
        return "{}&pageOffset={}&pageSize={}".format(url, self.page, self.page_size)

    def has_field(self, field):
        """
        is filter has field

        :param str field: field to check
        :return: Ture is have it
        :rtype: bool
        """
        for single_filter in self.or_filter_list:
            if single_filter.field == field:
                return True

        for single_filter in self.and_filter_list:
            if single_filter.field == field:
                return True

        return False

    def pop(self, field):
        """
        Pop filed

        :param str field: field to pop
        """
        for single_filter in self.or_filter_list:
            if single_filter.field == field:
                self.or_filter_list.remove(single_filter)

        for single_filter in self.and_filter_list:
            if single_filter.field == field:
                self.and_filter_list.remove(single_filter)

    def pop_join(self, field):
        """
        Pop join

        :param str field: field to pop
        """
        if self.join is not None:
            for single_filter in self.join["filter"]["$and"]:
                if field in single_filter:
                    self.join["filter"]["$and"].remove(single_filter)

    def add_join(self, field, values, operator: FiltersOperations = None, method: FiltersMethod = FiltersMethod.AND):
        """
        join a query to the filter

        :param str field: Metadata field / attribute
        :param str or list values: field values
        :param dl.FiltersOperations operator: optional - in, gt, lt, eq, ne
        :param method: optional - str - FiltersMethod.AND, FiltersMethod.OR

        **Example**:

        .. code-block:: python

            filter.add_join(field='metadata.user', values=['1','2'], operator=dl.FiltersOperations.IN)
        """
        if self.resource not in [FiltersResource.ITEM, FiltersResource.ANNOTATION]:
            raise exceptions.PlatformException(error="400", message="Cannot join to {} filters".format(self.resource))

        if self.join is None:
            self.join = dict()
        if "on" not in self.join:
            if self.resource == FiltersResource.ITEM:
                self.join["on"] = {"resource": FiltersResource.ANNOTATION.value, "local": "itemId", "forigen": "id"}
            else:
                self.join["on"] = {"resource": FiltersResource.ITEM.value, "local": "id", "forigen": "itemId"}
        if "filter" not in self.join:
            self.join["filter"] = dict()
        join_method = "$" + method
        if join_method not in self.join["filter"]:
            self.join["filter"][join_method] = list()
        self.join["filter"][join_method].append(SingleFilter(field=field, values=values, operator=operator).prepare())

    def __add_defaults(self):
        if self._use_defaults:
            # add items defaults
            if self.resource == FiltersResource.ITEM:
                self._unique_fields = ["type", "hidden"]
                self.add(field="hidden", values=False, method=FiltersMethod.AND)
                self.add(field="type", values="file", method=FiltersMethod.AND)
            # add service defaults
            elif self.resource == FiltersResource.SERVICE:
                self._unique_fields = ["global"]
                self.add(field="global", values=True, operator=FiltersOperations.NOT_EQUAL, method=FiltersMethod.AND)
            elif self.resource == FiltersResource.PACKAGE:
                self._unique_fields = ["global"]
                self.add(field="global", values=True, operator=FiltersOperations.NOT_EQUAL, method=FiltersMethod.AND)
            # add annotations defaults
            elif self.resource == FiltersResource.ANNOTATION:
                self._unique_fields = ["type"]
                values = [annotation_type.value for annotation_type in entities.AnnotationType]
                values.remove(entities.AnnotationType.NOTE.value)
                self.add(field="type", values=values, operator=FiltersOperations.IN, method=FiltersMethod.AND)

    def __generate_query(self):
        filters_dict = dict()

        if len(self.or_filter_list) > 0:
            or_filters = list()
            for single_filter in self.or_filter_list:
                or_filters.append(
                    single_filter.prepare(recursive=self.recursive and self.resource == FiltersResource.ITEM)
                )
            filters_dict["$or"] = or_filters

        if len(self.and_filter_list) > 0:
            and_filters = list()
            for single_filter in self.and_filter_list:
                and_filters.append(
                    single_filter.prepare(recursive=self.recursive and self.resource == FiltersResource.ITEM)
                )
            filters_dict["$and"] = and_filters

        return filters_dict

    def __generate_custom_query(self):
        if "filter" not in self.custom_filter:
            query_dict = {"filter": self.custom_filter}
        else:
            query_dict = self.custom_filter
        if "resource" not in query_dict:
            query_dict["resource"] = self.resource
        if "page" not in query_dict:
            query_dict["page"] = self.page
        if "pageSize" not in query_dict:
            query_dict["pageSize"] = self.page_size
        if self.join is not None and 'join' not in query_dict:
            query_dict["join"] = self.join
        if "join" in query_dict and "on" not in query_dict["join"]:
            if self.resource == FiltersResource.ITEM:
                query_dict["join"]["on"] = {
                    "resource": FiltersResource.ANNOTATION.value,
                    "local": "itemId",
                    "forigen": "id",
                }
            else:
                query_dict["join"]["on"] = {"resource": FiltersResource.ITEM.value, "local": "id", "forigen": "itemId"}

        return query_dict

    def __generate_ref_query(self):
        refs = list()
        if self._ref_task:
            task_refs = list()
            if not isinstance(self._ref_task_id, list):
                self._ref_task_id = [self._ref_task_id]

            for ref_id in self._ref_task_id:
                task_refs.append({"type": "task", "id": ref_id})

            refs += task_refs

        if self._ref_assignment:
            assignment_refs = list()
            if not isinstance(self._ref_assignment_id, list):
                self._ref_assignment_id = [self._ref_assignment_id]

            for ref_id in self._ref_assignment_id:
                assignment_refs.append({"type": "assignment", "id": ref_id})

            refs += assignment_refs

        return refs

    def prepare(self, operation=None, update=None, query_only=False, system_update=None, system_metadata=False):
        """
        To dictionary for platform call

        :param str operation: operation
        :param update: update
        :param bool query_only: query only
        :param system_update: system update
        :param system_metadata: True, if you want to change metadata system
        :return: dict of the filter
        :rtype: dict
        """
        ########
        # json #
        ########
        _json = dict()

        if self.custom_filter is not None:
            _json = self.__generate_custom_query()
            return _json

        _json["filter"] = self.__generate_query()

        ##################
        # filter options #
        ##################
        if not query_only:
            if len(self.sort) > 0:
                _json["sort"] = self.sort

            self.__validate_page_size()

            _json["page"] = self.page
            _json["pageSize"] = self.page_size
            _json["resource"] = self.resource

        ########
        # join #
        ########
        if self.join is not None:
            _json["join"] = self.join

        #####################
        # operation or refs #
        #####################
        if self._ref_assignment or self._ref_task:
            _json["references"] = {"operation": self._ref_op, "refs": self.__generate_ref_query()}
        elif operation is not None:
            if operation == "update":
                if update:
                    _json[operation] = {"metadata": {"user": update}}
                else:
                    _json[operation] = dict()
                if system_metadata and system_update:
                    _json["systemSpace"] = True
                    _json[operation]["metadata"] = _json[operation].get("metadata", dict())
                    _json[operation]["metadata"]["system"] = system_update
            elif operation == "delete":
                _json[operation] = True
                _json.pop("sort", None)
                if self.resource == FiltersResource.ITEM:
                    _json.pop("page", None)
                    _json.pop("pageSize", None)
            else:
                raise exceptions.PlatformException(error="400", message="Unknown operation: {}".format(operation))

        if self.context is not None:
            _json["context"] = self.context
        if self._system_space is not None:
            _json["systemSpace"] = self._system_space
        return _json

    def print(self, indent=2):
        print(json.dumps(self.prepare(), indent=indent))

    def sort_by(self, field, value: FiltersOrderByDirection = FiltersOrderByDirection.ASCENDING):
        """
        sort the filter

        :param str field: field to sort by it
        :param dl.FiltersOrderByDirection value: FiltersOrderByDirection.ASCENDING, FiltersOrderByDirection.DESCENDING

        **Example**:

        .. code-block:: python

            filter.sort_by(field='metadata.user', values=dl.FiltersOrderByDirection.ASCENDING)
        """
        if value not in [FiltersOrderByDirection.ASCENDING, FiltersOrderByDirection.DESCENDING]:
            raise exceptions.PlatformException(error="400", message="Sort can be by ascending or descending order only")
        self.sort[field] = value.value if isinstance(value, FiltersOrderByDirection) else value

    def platform_url(self, resource) -> str:
        """
        Build a url with filters param to open in web browser

        :param str resource: dl entity to apply filter on. currently only supports dl.Dataset
        :return: url string
        :rtype: str
        """
        _json = self.prepare()
        # add the view option
        _json["view"] = "icons"
        # convert from enum to string
        _json["resource"] = f'{_json["resource"]}'
        # convert the dictionary to a json string
        _json["dqlFilter"] = json.dumps(
            {"filter": _json.pop("filter"), "join": _json.pop("join", None), "sort": _json.get("sort", None)}
        )
        # set the page size as the UI default
        _json["pageSize"] = 100
        _json["page"] = _json["page"]
        # build the url for the dataset data browser
        if isinstance(resource, entities.Dataset):
            url = resource.platform_url + f"?{urllib.parse.urlencode(_json)}"
        else:
            raise NotImplementedError("Not implemented for resource type: {}".format(type(resource)))
        return url

    def open_in_web(self, resource):
        """
        Open the filter in the platform data browser (in a new web browser)

        :param str resource: dl entity to apply filter on. currently only supports dl.Dataset
        """
        if isinstance(resource, entities.Dataset):
            resource._client_api._open_in_web(url=self.platform_url(resource=resource))
        else:
            raise NotImplementedError("Not implemented for resource type: {}".format(type(resource)))

    def save(self, project: entities.Project, filter_name: str):
        """
        Save the current DQL filter to the project

        :param project: dl.Project
        :param filter_name: the saved filter's name
        :return: True if success
        """
        _json_filter = self.prepare()
        shebang_dict = {
            "type": "dql",
            "shebang": "dataloop",
            "metadata": {
                "version": "1.0.0",
                "system": {"mimetype": "dql"},
                "dltype": "filter",
                "filterFieldsState": [],
                "resource": "items",
                "filter": _json_filter.pop("filter"),
                "join": _json_filter.pop("join"),
            },
        }
        b_dataset = project.datasets._get_binaries_dataset()
        byte_io = io.BytesIO()
        byte_io.name = filter_name
        byte_io.write(json.dumps(shebang_dict).encode())
        byte_io.seek(0)
        b_dataset.items.upload(local_path=byte_io, remote_path="/.dataloop/dqlfilters/items", remote_name=filter_name)
        return True

    @classmethod
    def load(cls, project: entities.Project, filter_name: str) -> "Filters":
        """
        Load a saved filter from the project by name

        :param project: dl.Project entity
        :param filter_name: filter name
        :return: dl.Filters
        """
        b_dataset = project.datasets._get_binaries_dataset()
        f = entities.Filters(
            custom_filter={
                "filter": {"$and": [{"filename": f"/.dataloop/dqlfilters/items/{filter_name}"}]},
                "page": 0,
                "pageSize": 1000,
                "resource": "items",
            }
        )
        pages = b_dataset.items.list(filters=f)
        if pages.items_count == 0:
            raise exceptions.NotFound(
                f"Saved filter not found: {filter_name}. Run `Filters.list()` to list existing filters"
            )
        with open(pages.items[0].download()) as f:
            data = json.load(f)
            custom_filter = data["metadata"]["filter"]
            custom_filter["join"] = data["metadata"]["join"]
        return cls(custom_filter=custom_filter)

    @staticmethod
    def list(project: entities.Project) -> list:
        """
        List all saved filters for a project
        :param project: dl.Project entity
        :return: a list of all the saved filters' names
        """
        b_dataset = project.datasets._get_binaries_dataset()
        f = entities.Filters(use_defaults=False, field="dir", values="/.dataloop/dqlfilters/items")
        pages = b_dataset.items.list(filters=f)
        all_filter_items = list(pages.all())
        names = [i.name for i in all_filter_items]
        return names

    @staticmethod
    def _get_split_filters(dataset, filters, max_items, max_workers=4, max_depth=None) -> Generator[dict, None, None]:
        """
        Generator that yields filter chunks for large datasets using a bounded
        thread pool. Splits ranges by id until each subset holds <= max_items.

        :param dataset: Dataset object to get filters for
        :param filters: Base filters to apply
        :param max_items: Maximum number of items per filter chunk
        :param max_workers: Maximum number of threads for parallel processing
        :param max_depth: Maximum depth of the filter tree. Default calculated by the formula: np.ceil(np.log2(count/max_items) + 3).
        :yield: Filter payloads covering subsets of items
        """
        if max_items <= 0:
            raise ValueError("_get_split_filters : max_items must be greater than 0")

        if filters is None:
            filters = entities.Filters()

        from_id, count = Filters._get_first_last_item(
            items_repo=dataset.items, filters=filters, order_by_direction=FiltersOrderByDirection.ASCENDING
        )
        to_id, count = Filters._get_first_last_item(
            items_repo=dataset.items, filters=filters, order_by_direction=FiltersOrderByDirection.DESCENDING
        )

        if from_id is None or to_id is None or count == 0:
            return

        max_depth = max_depth if max_depth is not None else np.ceil(np.log2(count / max_items) + 3)

        def make_filter_dict(range_from_id, range_to_id, strict_from: bool = False):
            fdict = copy.deepcopy(filters.prepare())
            lower_op = "$gt" if strict_from else "$gte"
            fdict["filter"].setdefault("$and", []).extend(
                [{"id": {lower_op: range_from_id}}, {"id": {"$lte": range_to_id}}]
            )
            return fdict

        def task(range_from_id, range_to_id, depth, strict_from: bool):
            fdict = make_filter_dict(range_from_id, range_to_id, strict_from)
            range_filters = entities.Filters(custom_filter=fdict, page_size=1)
            actual_from, count = Filters._get_first_last_item(
                dataset.items, range_filters, FiltersOrderByDirection.ASCENDING
            )
            if count == 0:
                return ("none", None, None)
            if count <= max_items or depth >= max_depth:
                return ("yield", fdict, None)
            actual_to, count = Filters._get_first_last_item(
                dataset.items, range_filters, FiltersOrderByDirection.DESCENDING
            )
            if not actual_from or not actual_to or actual_from == actual_to:
                return ("yield", fdict, None)
            mid = Filters._get_middle_id(actual_from, actual_to)
            if not mid or mid == actual_from or mid == actual_to:
                return ("yield", fdict, None)
            # Left child: [actual_from, mid] inclusive; Right child: (mid, actual_to] exclusive lower bound
            return (
                "split",
                None,
                (
                    (actual_from, mid, depth + 1, False),  # left child includes lower bound
                    (mid, actual_to, depth + 1, True),  # right child excludes midpoint
                ),
            )

        pending = deque([(from_id, to_id, 0, False)])
        futures = set()

        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            while futures or pending:
                # Submit all pending tasks
                while pending:
                    fr, to, d, strict = pending.popleft()
                    futures.add(pool.submit(task, fr, to, d, strict))

                if not futures:
                    break

                done, futures = wait(futures, return_when=FIRST_COMPLETED)
                for fut in done:
                    try:
                        kind, fdict, ranges = fut.result()
                    except Exception as e:
                        logger.warning(f"split filters task failed: {e}")
                        continue
                    if kind == "yield" and fdict is not None:
                        yield fdict
                    elif kind == "split" and ranges is not None:
                        left, right = ranges
                        pending.append(left)
                        pending.append(right)

    @staticmethod
    def _get_first_last_item(
        items_repo, filters, order_by_direction=FiltersOrderByDirection.ASCENDING
    ) -> Tuple[Optional[str], int]:
        filters_dict = copy.deepcopy(filters.prepare())
        filters_dict["sort"] = {"id": order_by_direction.value}
        filters_dict["page"] = 0
        filters_dict["pageSize"] = 1
        cloned_filters = entities.Filters(custom_filter=filters_dict)

        try:
            pages = items_repo.list(filters=cloned_filters)
            return (pages.items[0].id if pages.items else None, pages.items_count)
        except Exception:
            return None, 0

    @staticmethod
    def _get_middle_id(from_id, to_id):
        """Calculate middle ObjectId between two ObjectIds with sub-second precision.

        Computes the midpoint in the full 12-byte ObjectId numeric space to avoid
        second-level rounding inherent to datetime-based construction.
        """
        try:
            # Convert ObjectId strings to integers using base 16 (hexadecimal)
            start_int = int(str(ObjectId(from_id)), base=16)  
            end_int = int(str(ObjectId(to_id)), base=16)
            if start_int >= end_int:
                return from_id
            mid_int = (start_int + end_int) // 2
            if mid_int <= start_int:
                mid_int = start_int + 1
            if mid_int > end_int:
                mid_int = end_int
            # Convert back to 12-byte ObjectId format
            mid_bytes = mid_int.to_bytes(length=12, byteorder="big")
            return str(ObjectId(mid_bytes))
        except Exception:
            return from_id  # Fallback to from_id if calculation fails


class SingleFilter:
    def __init__(self, field, values, operator: FiltersOperations = None):
        self.field = field
        self.values = values
        self.operator = operator

    @staticmethod
    def __add_recursive(value):
        if not value.endswith("*") and not os.path.splitext(value)[-1].startswith("."):
            if value.endswith("/"):
                value = value + "**"
            else:
                value = value + "/**"
        return value

    def prepare(self, recursive=False):
        """
        To dictionary for platform call

        :param recursive:recursive
        """
        _json = dict()
        values = self.values

        if recursive and self.field == "filename":
            if isinstance(values, str):
                values = self.__add_recursive(value=values)
            elif isinstance(values, list):
                for i_value, value in enumerate(values):
                    values[i_value] = self.__add_recursive(value=value)

        if self.operator is None:
            _json[self.field] = values
        else:
            value = dict()
            op = self.operator.value if isinstance(self.operator, FiltersOperations) else self.operator
            value["${}".format(op)] = values
            _json[self.field] = value

        return _json

    def print(self, indent=2):
        print(json.dumps(self.prepare(), indent=indent))



import datetime

import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def first_method(self, dataset, progress=None):
        """
        Write your main package service here

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        assert isinstance(dataset, dl.entities.Dataset)
        filters = dl.Filters(field='annotated', values=True)
        filters.add_join(field='label', values='person')
        task = dataset.tasks.create(task_name='AutomatedTask',
                                    due_date=datetime.datetime.now().timestamp() + 60 * 60 * 24 * 7,
                                    assignee_ids=['annotator1@dataloop.ai', 'annotator2@dataloop.ai'])
        logger.info('Task created successfully. Task name: {}'.format(task.name))

    def second_method(self, dataset, progress=None):
        """
        Write your main package service here

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        assert isinstance(dataset, dl.entities.Dataset)


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



def main(project_name, dataset_name, remote_filepath):
    """
    Copy annotations between items
    :return:
    """
    import dtlpy as dl

    # FROM get the annotations from item
    project = dl.projects.get(project_name=project_name)
    dataset = project.datasets.get(dataset_name=dataset_name)
    item = dataset.items.get(filepath=remote_filepath)

    # get annotations
    annotations = item.annotations.list()

    # delete first annotation
    ann = annotations[0]
    ann.delete()

    ### Or - to delete all annotations ###

    # get annotations
    annotations = item.annotations.list()

    # delete first annotation
    annotations.delete()



def main():
    """
    Convert annotation types
    :return:
    """
    from PIL import Image
    import matplotlib.pyplot as plt
    import dtlpy as dl

    # Get project and dataset
    project = dl.projects.get(project_name='Toilet Paper')
    dataset = project.datasets.get(dataset_name='Extra Soft')

    # Get item and binaries
    item = dataset.items.get(filepath='/with_puppies.jpg')
    buffer = item.download(save_locally=False)

    #######################################
    # Convert mask annotations to polygon #
    #######################################
    img = Image.open(buffer)
    builder = item.annotations.builder()
    builder.add(dl.Polygon.from_segmentation(mask=mask,  # binary mask of the annotation
                                             label='roll'))
    # plot the annotation
    plt.figure()
    plt.imshow(builder.show())
    # plot annotation on the image
    plt.figure()
    plt.imshow(builder.show(image=img))

    # upload annotation to platform
    item.annotations.update(builder)

    #######################################
    # Convert polygon annotations to mask #
    #######################################
    img = Image.open(buffer)
    builder = item.annotations.builder()
    builder.add(dl.Segmentation.from_polygon(geo=[[x1, y1], [x2, y2], [x3, y3]],  # list of coordinates
                                             shape=img.size[::-1],  # (h,w)
                                             label='roll'))
    # plot the annotation
    plt.figure()
    plt.imshow(builder.show())
    # plot annotation on the image
    plt.figure()
    plt.imshow(builder.show(image=img))

    # upload annotation to platform
    item.annotations.update(builder)



"""
Messages Repository
"""
import json
import logging
from urllib.parse import urlencode, quote
from .. import entities, miscellaneous, exceptions, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Messages:
    """
    Messages Repository

    The Messages class allows the user to manage Messages.
    """

    def __init__(self, client_api: ApiClient):
        self._client_api = client_api

    # @_api_reference.add(path='/inbox/message/user', method='get')
    def _list(self, context: entities.NotificationEventContext = None, checkpoint: dict = None, new_only=True) -> \
            miscellaneous.List[entities.Message]:
        """
        Get messages by context.

        :param str context: Message context
        :param dict checkpoint: checkpoint
        :param bool new_only: get only new messages
        :return: List of Messages
        :rtype: list

        **Example**:

        .. code-block:: python

            messages = dl.messages.list(context={project: id})
        """

        user = self._client_api.info()['user_email']
        if not user:
            raise exceptions.PlatformException(error='400',
                                               message='No user in JWT, please login')

        url = '/inbox/message/user'

        query_params = {
            'newOnly': new_only,
            'checkpointPagination': {
                "pageSize": 1000,
                "sort": {"key": "id", "direction": "descending"}
            }
        }

        context_extent = None

        if context:
            context_extent = "contextExtent=" + quote(json.dumps(context).replace(" ", ""), safe=":,").strip()

        if checkpoint:
            query_params['checkpointPagination']['checkpoint'] = checkpoint

        checkpoint_pagination = "checkpointPagination=" + quote(
            json.dumps(query_params['checkpointPagination']).replace(" ", ""), safe=":,").strip()

        url = "{}?{}".format(url, checkpoint_pagination)

        if context_extent:
            url = "{}&{}".format(url, context_extent)

        success, response = self._client_api.gen_request(req_type='get', path=url)

        if success:
            pool = self._client_api.thread_pools('entity.create')
            message_json = response.json()
            items = message_json.get('items', list())
            jobs = [None for _ in range(len(items))]
            for i_message, message in enumerate(items):
                jobs[i_message] = pool.submit(
                    entities.Message._protected_from_json,
                    **{'client_api': self._client_api, '_json': message}
                )

            # get all results
            results = [j.result() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            messages = miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            raise exceptions.PlatformException(response)
        return messages



import logging
import os
import zipfile
from typing import List

import numpy as np
import pathspec

logger = logging.getLogger(name='dtlpy')

MAX_ZIP_FILE = 100e6  # 100MB


class Zipping:
    def __init__(self):
        pass

    @staticmethod
    def zip_directory(zip_filename, directory=None, ignore_max_file_size=False, ignore_directories: List[str] = None):
        """
        Zip Directory
        Will ignore .gitignore files

        :param directory: the directory to zip
        :param zip_filename: the name of the zipfile.
        :param ignore_max_file_size: ignore the limitation on the zip file size
        :param list[str] ignore_directories: directories to ignore.
        :return: None
        """
        # default path
        if directory is None:
            directory = os.getcwd()
        # check if directory
        assert os.path.isdir(directory), '[ERROR] Directory does not exists: {}'.format(directory)

        if '.gitignore' in os.listdir(directory):
            with open(os.path.join(directory, '.gitignore')) as f:
                spec_src = f.read()
        else:
            spec_src = ''
        ignore_lines = spec_src.splitlines() + ['.git', '.dataloop']
        if ignore_directories is not None:
            ignore_lines += ignore_directories
        spec = pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, ignore_lines)

        # init zip file
        zip_file = zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED)
        try:
            for root, dirs, files in os.walk(directory):
                # remove dirs to avoid going file by file
                for d in dirs:
                    if spec.match_file(os.path.relpath(os.path.join(root, d), directory)):
                        dirs.remove(d)
                for file in files:
                    filepath = os.path.join(root, file)
                    if not spec.match_file(os.path.relpath(filepath, directory)):
                        Zipping.__add_to_zip_file(directory, filepath, ignore_max_file_size, zip_file)
        finally:
            zip_file.close()

    @staticmethod
    def zip_directory_inclusive(zip_filename, directory=None, ignore_max_file_size=False,
                                subpaths: List[str] = None):
        """
        Zip Directory
        Will ignore .gitignore files

        :param directory: the directory to zip.
        :param zip_filename: the name of the zipfile
        :param ignore_max_file_size: ignore the limitation on the zip file size
        :param list[str] subpaths: paths to include in the final zip (relative path).
        :return: None
        """
        # default path
        if directory is None:
            directory = os.getcwd()
        # check if directory
        assert os.path.isdir(directory), '[ERROR] Directory does not exists: %s' % directory

        if '.gitignore' in os.listdir(directory):
            with open(os.path.join(directory, '.gitignore')) as f:
                spec_src = f.read()
        else:
            spec_src = ''
        ignore_lines = spec_src.splitlines() + ['.git', '.dataloop']
        spec = pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, ignore_lines)

        # init zip file
        zip_file = zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED)
        try:
            for root, dirs, files in os.walk(directory):
                for file in files:
                    filepath = os.path.join(root, file)
                    if not spec.match_file(os.path.relpath(filepath, directory)) \
                            and Zipping.__check_filepath(os.path.relpath(filepath, directory), subpaths):
                        Zipping.__add_to_zip_file(directory, filepath, ignore_max_file_size, zip_file)
        finally:
            zip_file.close()

    @staticmethod
    def __check_filepath(filepath: str, paths: List[str]):
        """
        Checks whether a specific file is inside one of the subdirectories
        """
        return any(filepath.startswith(directory) for directory in paths)

    @staticmethod
    def __add_to_zip_file(directory, filepath, ignore_max_file_size, zip_file):
        zip_file.write(filepath, arcname=os.path.relpath(filepath, directory))
        if not ignore_max_file_size:
            if np.sum([f.file_size for f in list(zip_file.NameToInfo.values())]) > MAX_ZIP_FILE:
                logger.error('Failed zipping in file: {}'.format(filepath))
                raise ValueError(
                    'Zip file cant be over 100MB. '
                    'Please verify that only code is being uploaded or '
                    'add files to .gitignore so they wont be zipped and uploaded as code.')

    @staticmethod
    def unzip_directory(zip_filename, to_directory=None):
        with zipfile.ZipFile(zip_filename) as zipdata:
            zipinfos = zipdata.infolist()
            # iterate through each file
            for zipinfo in zipinfos:
                # encode the file names
                # zip package make decode by cp437 for file that have name that not ascii
                # this happen when the flag_bits be different than 0
                # so we encode the name back
                if not zipinfo.flag_bits:
                    zipinfo.filename = zipinfo.filename.encode('cp437').decode('utf-8')
                zipdata.extract(zipinfo, to_directory)



import logging

from .. import entities, exceptions, miscellaneous, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Apps:

    def __init__(self, client_api: ApiClient, project: entities.Project = None, project_id: str = None):
        self._client_api = client_api
        self._project = project
        self._project_id = project_id 
        self._commands = None

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._project_id is None:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Missing "project". need to set a Project entity or use project.apps repository')
            else:
                self._project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def commands(self) -> repositories.Commands:
        if self._commands is None:
            self._commands = repositories.Commands(client_api=self._client_api)
        return self._commands

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def get(self,
            app_name: str = None,
            app_id: str = None,
            fetch: bool = None) -> entities.App:
        """
        Get an app object.

        note: It's required to pass either app_id of app_name

        :param str app_id: optional - search by id.
        :param str app_name: optional - search by name.
        :param bool fetch: optional - fetch entity from platform, default taken from cookie

        **Example**:

        .. code-block:: python
            app = self.apps.get(app_id='app_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if app_id is None and app_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='You must provide an identifier in inputs')
        if fetch:
            if app_name is not None:
                app = self.__get_by_name(name=app_name)
            else:
                success, response = self._client_api.gen_request(req_type='get', path="/apps/{}".format(app_id))
                if not success:
                    raise exceptions.PlatformException(response)
                app = entities.App.from_json(client_api=self._client_api,
                                             _json=response.json(),
                                             project=self._project)
        else:
            app = entities.App.from_json(
                _json={
                    'id': app_id,
                    'name': app_name
                },
                client_api=self._client_api,
                project=self._project,
                is_fetched=False
            )
        assert isinstance(app, entities.App)
        return app

    def __get_by_name(self, name) -> entities.App:
        filters = entities.Filters(field='name',
                                   values=name,
                                   resource=entities.FiltersResource.APP,
                                   use_defaults=False)
        if self._project is not None:
            filters.add(field='projectId', values=self.project.id)
        apps = self.list(filters=filters)
        if apps.items_count == 0:
            raise exceptions.PlatformException(
                error='404',
                message='app not found. Name: {}'.format(name))
        elif apps.items_count > 1:
            raise exceptions.PlatformException(
                error='400',
                message='More than one app found by the name of: {} '.format(apps))
        return apps.items[0]

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.App]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.App._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item,
                                          'project': self._project})
        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items

    def _list(self, filters: entities.Filters):
        url = '/apps/query'

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None, project_id: str = None) -> entities.PagedEntities:
        """
        List the available apps in the specified project.

        :param entities.Filters filters: the filters to apply to the list
        :param str project_id: the project id to apply thew filters on.
        :return a paged entity representing the list of apps.

        ** Example **
        .. code-block:: python
            apps = dl.apps.list(project_id='id')
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.APP)
            # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.APP:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.APP. Got: {!r}'.format(filters.resource))

        # noinspection DuplicatedCode
        if project_id is None and self._project is not None:
            project_id = self._project.id

        if project_id is not None:
            filters.add(field='projectId', values=project_id)

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       project_id=project_id,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def update(self, app: entities.App = None, app_id: str = None, wait: bool = True) -> bool:
        """
        Update the current app to the new configuration

        :param entities.App app: The app to update.
        :param str app_id: The app id to update.
        :param bool wait: wait for the operation to finish.
        :return bool whether the operation ran successfully or not

        **Example**
        .. code-block:: python
            succeed = dl.apps.update(app)
        """
        if app_id is not None and app is None:
            app = self.get(app_id=app_id)
        if app is None:
            raise exceptions.PlatformException(error='400', message='You must provide app or app_id')
        success, response = self._client_api.gen_request(req_type='put',
                                                         path=f"/apps/{app.id}",
                                                         json_req=app.to_json())
        if not success:
            raise exceptions.PlatformException(response)

        app = entities.App.from_json(
            _json=response.json(),
            client_api=self._client_api,
            project=self.project
        )
        if app.metadata:
            command_id = app.metadata.get('system', {}).get('commands', {}).get('update', None)
            if wait and app.status == entities.CompositionStatus.UPDATING and command_id is not None:
                command = self.commands.get(command_id=command_id, url='api/v1/commands/faas/{}'.format(command_id))
                command.wait()
                app = self.get(app_id=app.id)

        return success

    def install(self,
                dpk: entities.Dpk,
                app_name: str = None,
                organization_id: str = None,
                custom_installation: dict = None,
                scope: entities.AppScope = None,
                wait: bool = True,
                integrations: list = None
                ) -> entities.App:
        """
        Install the specified app in the project.

        Note: You must pass either the app_id or app_name
        :param entities.App dpk: the app entity
        :param str app_name: installed app name. default is the dpk name
        :param str organization_id: the organization which you want to apply on the filter.
        :param dict custom_installation: partial installation.
        :param str scope: the scope of the app. default is project.
        :param bool wait: wait for the operation to finish.
        :param list integrations: list of integrations to install with the app.

        :return the installed app.
        :rtype entities.App

        **Example**
        .. code-block:: python
            app = dl.apps.install(dpk=dpk)
        """
        if dpk is None:
            raise exceptions.PlatformException(error='400', message='You must provide an app')

        if app_name is None:
            app_name = dpk.display_name
        if isinstance(scope, entities.AppScope):
            scope = scope.value
        app = entities.App.from_json(_json={'name': app_name,
                                            'projectId': self.project.id,
                                            'orgId': organization_id,
                                            'dpkName': dpk.name,
                                            "customInstallation": custom_installation,
                                            'dpkVersion': dpk.version,
                                            'scope': scope,
                                            'integrations': integrations
                                            },
                                     client_api=self._client_api,
                                     project=self.project)
        success, response = self._client_api.gen_request(req_type='post',
                                                         path="/apps",
                                                         json_req=app.to_json())
        if not success:
            raise exceptions.PlatformException(response)
        app = entities.App.from_json(_json=response.json(),
                                     client_api=self._client_api,
                                     project=self.project)

        if app.metadata:
            command_id = app.metadata.get('system', {}).get('commands', {}).get('install', None)
            if wait and app.status == entities.CompositionStatus.INITIALIZING and command_id is not None:
                command = self.commands.get(command_id=command_id, url='api/v1/commands/faas/{}'.format(command_id))
                command.wait()
                app = self.get(app_id=app.id)

        return app

    def uninstall(self, app_id: str = None, app_name: str = None, wait: bool = True, app: entities.App = None) -> bool:
        """
        Delete an app entity.

        Note: You are required to add either app_id or app_name.

        :param str app_id: optional - the id of the app.
        :param str app_name: [DEPRECATED] - the name of the app.
        :param bool wait: optional - wait for the operation to finish.
        :param entities.App app: optional - the app entity.
        :return whether we succeed uninstalling the specified app.
        :rtype bool

        **Example**
        .. code-block:: python
            # succeed = dl.apps.delete(app_id='app_id')
        """
        if app is None and app_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='You must provide an identifier in inputs')

        if app is not None:
            app_id = app.id

        success, response = self._client_api.gen_request(req_type='delete', path='/apps/{}'.format(app_id))
        if not success:
            raise exceptions.PlatformException(response)

        try:
            app = self.get(app_id=app_id)
        except exceptions.NotFound:
            return success

        if wait and app.status == entities.CompositionStatus.TERMINATING and app.metadata is not None:
            command_id = app.metadata.get('system', {}).get('commands', {}).get('uninstall', None)
            if command_id is not None:
                command = self.commands.get(command_id=command_id, url='api/v1/commands/faas/{}'.format(command_id))
                command.wait()

        logger.debug(f"App deleted successfully (id: {app.id}, name: {app.name}")
        return success

    def resume(self, app: entities.App = None, app_id: str = None) -> bool:
        """
        Activate an app entity.

        Note: You are required to add either app or app_id.

        :param entities.App app: the app entity
        :param str app_id: optional - the id of the app.
        :return whether we succeed activating the specified app.
        :rtype bool

        **Example**
        .. code-block:: python
            # succeed = dl.apps.resume(app)
        """
        if app_id is not None and app is None:
            app = self.get(app_id=app_id)

            if app and app.status == entities.CompositionStatus.INSTALLED:
                raise exceptions.PlatformException(
                    error='400',
                    message='Application is already active'
                )
        if app is None:
            raise exceptions.PlatformException(error='400', message='You must provide app or app_id')

        success, response = self._client_api.gen_request(req_type='post', path='/apps/{}/activate'.format(app.id))
        if not success:
            raise exceptions.PlatformException(response)

        logger.debug(f"App resumed successfully (id: {app.id}, name: {app.name}")
        return success

    def pause(self, app: entities.App = None, app_id: str = None) -> bool:
        """
        Pausing an app entity.

        Note: You are required to add either app or app_id.

        :param entities.App app: the app entity
        :param str app_id: optional - the id of the app.
        :return whether we succeed pausing the specified app.
        :rtype bool

        **Example**
        .. code-block:: python
            # succeed = dl.apps.pause(app)
        """
        if app_id is not None and app is None:
            app = self.get(app_id=app_id)

            if app and app.status == entities.CompositionStatus.UNINSTALLED:
                raise exceptions.PlatformException(
                    error='400',
                    message='Application is already inactive'
                )
        if app is None:
            raise exceptions.PlatformException(error='400', message='You must provide app or app_id')

        success, response = self._client_api.gen_request(req_type='post', path='/apps/{}/deactivate'.format(app.id))
        if not success:
            raise exceptions.PlatformException(response)

        logger.debug(f"App paused successfully (id: {app.id}, name: {app.name}")
        return success



import asyncio
import time
import types

import numpy as np
import os
import logging
import dtlpy as dl
import shutil
from dtlpy import miscellaneous

logger = logging.getLogger(name='dtlpy')


##########
# Videos #
##########
class Videos:
    def __init__(self):
        pass

    @staticmethod
    def get_info(filepath):
        try:
            import ffmpeg
        except ImportError:
            logger.error(
                'Import Error! Cant import ffmpeg. '
                'Annotations operations will be limited. import manually and fix errors')
            raise
        probe = ffmpeg.probe(filepath)
        return probe

    @staticmethod
    def get_max_object_id(item):
        max_object_id = 1
        annotations_list = item.annotations.list().annotations
        if len(annotations_list) < 1:
            return 1
        for annotation in annotations_list:
            if annotation.object_id is not None:
                current_object_id = int(annotation.object_id)
                if current_object_id > max_object_id:
                    max_object_id = current_object_id
        return max_object_id

    @staticmethod
    def video_snapshots_generator(item_id=None, item=None, frame_interval=30, image_ext="png"):
        futures = Videos._async_video_snapshots_generator(item_id=item_id,
                                                          item=item,
                                                          frame_interval=frame_interval,
                                                          image_ext=image_ext)
        loop = asyncio.new_event_loop()
        try:
            asyncio.set_event_loop(loop)
            return loop.run_until_complete(futures)
        finally:
            try:
                loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                asyncio.set_event_loop(None)
                loop.close()

    @staticmethod
    async def _async_video_snapshots_generator(item_id=None, item=None, frame_interval=30, image_ext="png"):
        """
        Create video-snapshots

        :param item_id: item id for the video
        :param item: item id for the video
        :param frame_interval: number of frames to take next snapshot
        :param image_ext: png/jpg
        :return: the uploaded items
        """
        if item_id is not None:
            item = dl.items.get(item_id=item_id)

        if item is None:
            raise ValueError('Missing input item (or item_id)')

        if not isinstance(frame_interval, int):
            raise AttributeError('frame_interval is mast to be integer')

        if "video" not in item.mimetype:
            raise AttributeError("Got {} file type but only video files are supported".format(item.mimetype))

        video_path = item.download()

        # Get the time for single frame from metadata (duration/# of frames)
        if 'system' in item.metadata and \
                'ffmpeg' in item.metadata['system'] and \
                'duration' in item.metadata['system']['ffmpeg'] and \
                'nb_frames' in item.metadata['system']['ffmpeg']:
            nb_frames = int(item.metadata["system"]["ffmpeg"]["nb_frames"])
            duration = float(item.metadata["system"]["ffmpeg"]["duration"])
            video_fps = duration / nb_frames
        else:
            try:
                import cv2
            except (ImportError, ModuleNotFoundError):
                logger.error(
                    'Import Error! Cant import cv2. '
                    'Annotations operations will be limited. import manually and fix errors')
                raise

            video = cv2.VideoCapture(video_path)
            video_fps = video.get(cv2.CAP_PROP_FPS)
            nb_frames = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
            duration = video_fps * nb_frames

        images_path = Videos.disassemble(filepath=video_path, frame_interval=frame_interval, image_ext=image_ext)
        snapshots_items = list()
        try:
            # rename files
            images = []
            video_basename = os.path.basename(video_path)
            for f in os.listdir(images_path):
                images.append(f)
            for image in images:
                image_split_name, ext = os.path.splitext(image)
                try:
                    frame = int(image_split_name) * frame_interval
                    file_frame_name = "{}.frame.{}{}".format(video_basename, frame, ext)
                    full_path = os.path.join(images_path, file_frame_name)
                    os.rename(os.path.join(images_path, image),
                              full_path)
                except Exception as e:
                    logger.debug("Rename {} has been failed: {}".format(os.path.join(images_path, image), e))

                remote_path = os.path.join(os.path.split(item.filename)[0], "snapshots")
                remote_url = '/items/{}/snapshots'.format(item.id)
                to_upload = open(full_path, 'rb')
                try:
                    response = await item._client_api.upload_file_async(to_upload=to_upload,
                                                                        item_type='file',
                                                                        item_size=os.stat(full_path).st_size,
                                                                        remote_url=remote_url,
                                                                        uploaded_filename=file_frame_name,
                                                                        remote_path=remote_path)
                except Exception:
                    raise
                finally:
                    to_upload.close()
                if response.ok:
                    snapshots_items.append(item.from_json(response.json(), item._client_api))
                else:
                    raise dl.PlatformException(response)

            # classification tpe annotation creation for each file
            builder = item.annotations.builder()
            annotation_itemlinks = []

            max_object_id = Videos().get_max_object_id(item=item)
            for snapshot_item in snapshots_items:
                max_object_id += 1
                item_frame = snapshot_item.name.rsplit("frame.", 1)[1].split(".")[0]
                if item_frame.isnumeric():
                    item_time = int(item_frame) * video_fps
                else:
                    item_frame = item_time = 0

                snapshot_item.metadata["system"]["itemLinks"] = [{"type": "snapshotFrom",
                                                                  "itemId": item.id,
                                                                  "frame": item_frame,
                                                                  "time": item_time}]

                annotation_itemlinks.append({"type": "snapshotTo",
                                             "itemId": snapshot_item.id,
                                             "frame": item_frame,
                                             "time": item_time})

                snapshot_item.update(system_metadata=True)
                annotation_definition = dl.Classification(label="Snapshot")
                builder.add(annotation_definition=annotation_definition,
                            frame_num=int(item_frame),
                            end_frame_num=nb_frames if int(item_frame) + int(video_fps) > nb_frames else int(
                                item_frame) + int(video_fps),
                            start_time=item_time,
                            end_time=duration if item_time + 1 > duration else item_time + 1,
                            object_id=max_object_id)

            annotations = item.annotations.upload(annotations=builder)

            # update system metadata for annotations
            count = 0
            for annotation in annotations:
                annotation.metadata["system"]["itemLinks"] = [annotation_itemlinks[count]]
                count += 1

            annotations.update(system_metadata=True)
        except Exception as err:
            logger.exception(err)
        finally:
            if os.path.isdir(images_path):
                shutil.rmtree(images_path)
            return snapshots_items

    @staticmethod
    def disassemble(filepath, fps=None, frame_interval=None, loglevel='panic', image_ext='jpg'):
        """
        Disassemble video to images

        :param filepath: input video filepath
        :param fps: rate of disassemble. e.g if 1 frame per second fps is 1. if None all frames will be extracted
        :param frame_interval: take image every frame # (if exists function ignore fps)
        :param image_ext: png/jpg
        :param loglevel: ffmpeg loglevel
        :return:
        """
        try:
            import ffmpeg
        except ImportError:
            logger.error(
                'Import Error! Cant import ffmpeg. '
                'Annotations operations will be limited. import manually and fix errors')
            raise
        # get video information
        video_props = Videos.get_info(filepath)
        if 'system' in video_props and \
                'nb_frames' in video_props['system'][0]:
            nb_frames = video_props['streams'][0]['nb_frames']
        else:
            try:
                import cv2
            except (ImportError, ModuleNotFoundError):
                logger.error(
                    'Import Error! Cant import cv2. '
                    'Annotations operations will be limited. import manually and fix errors')
                raise
            nb_frames = int(cv2.VideoCapture(filepath).get(cv2.CAP_PROP_FRAME_COUNT))

        if not os.path.isfile(filepath):
            raise IOError('File doesnt exists: {}'.format(filepath))
        basename, ext = os.path.splitext(filepath)
        # create folder for the frames
        if os.path.exists(basename):
            shutil.rmtree(basename)

        os.makedirs(basename, exist_ok=True)

        if fps is None:
            try:
                fps = eval(video_props['streams'][0]['avg_frame_rate'])
            except ZeroDivisionError:
                fps = 0
        num_of_zeros = len(str(nb_frames))
        # format the output filename
        output_regex = os.path.join(basename, '%0{}d.{}'.format(num_of_zeros, image_ext))

        try:
            if frame_interval is not None:
                frame_number = 0
                select = ""
                while frame_number < nb_frames:
                    if select != "":
                        select += '+'
                    select += 'eq(n\\,{})'.format(frame_number)
                    frame_number += frame_interval
                stream = ffmpeg.input(filepath, **{'loglevel': loglevel}).output(output_regex,
                                                                                 **{'start_number': '0',
                                                                                    'vf': 'select=\'{}'.format(select),
                                                                                    'vsync': 'vfr'})
            else:
                stream = ffmpeg.input(filepath, **{'loglevel': loglevel}).output(output_regex,
                                                                                 **{'start_number': '0',
                                                                                    'r': str(fps)})

            ffmpeg.overwrite_output(stream).run()
        except Exception:
            logger.error('ffmpeg error in disassemble:')
            raise
        return basename

    @staticmethod
    def reencode(filepath, loglevel='panic'):
        """
        Re-encode video as mp4, remove start offset and set bframes to 0

        :param filepath: input video file
        :param loglevel: ffmpeg loglevel
        :return:
        """
        try:
            import ffmpeg
        except ImportError:
            logger.error(
                'Import Error! Cant import ffmpeg. '
                'Annotations operations will be limited. import manually and fix errors')
            raise
        if not os.path.isfile(filepath):
            raise IOError('File doesnt exists: {}'.format(filepath))
        # re encode video without b frame and as mp4
        basename, ext = os.path.splitext(filepath)
        output_filepath = os.path.join(basename, os.path.basename(filepath).replace(ext, '.mp4'))
        if not os.path.isdir(os.path.dirname(output_filepath)):
            os.makedirs(os.path.dirname(output_filepath))
        try:
            stream = ffmpeg.input(filepath, **{'loglevel': loglevel}).output(output_filepath,
                                                                             **{'x264opts': 'bframes=0',
                                                                                'f': 'mp4'})
            ffmpeg.overwrite_output(stream).run()
        except Exception as e:
            logger.exception('ffmpeg error in disassemble:')
            raise

        output_probe = Videos.get_info(output_filepath)
        start_time = eval(output_probe['streams'][0]['start_time'])
        fps = eval(output_probe['streams'][0]['avg_frame_rate'])
        has_b_frames = output_probe['streams'][0]['has_b_frames']
        start_frame = fps * start_time
        if start_time != 0:
            logger.warning('Video start_time is not 0!')
        if has_b_frames != 0:
            logger.warning('Video still has b frames!')
        return output_filepath

    @staticmethod
    def split_and_upload(filepath,
                         # upload parameters
                         project_name=None, project_id=None, dataset_name=None, dataset_id=None, remote_path=None,
                         # split parameters
                         split_seconds=None, split_chunks=None, split_pairs=None,
                         loglevel='panic'):
        """
        Split video to chunks and upload to platform

        :param filepath: input video file
        :param project_name:
        :param project_id:
        :param dataset_name:
        :param dataset_id:
        :param remote_path:
        :param split_seconds: split by seconds per chunk. each chunk's length will be this in seconds
        :param split_chunks: split by number of chunks.
        :param split_pairs: a list od (start, stop) segments to split in seconds . e.g [(0,400), (600,800)]
        :param loglevel: ffmpeg loglevel
        :return:
        """
        try:
            import ffmpeg
        except ImportError:
            logger.error(
                'Import Error! Cant import ffmpeg. '
                'Annotations operations will be limited. import manually and fix errors')
            raise
        # https://www.ffmpeg.org/ffmpeg-formats.html#Examples-9

        # Validate filepath to prevent path traversal
        miscellaneous.PathUtils.validate_file_path(filepath)
        
        if not os.path.isfile(filepath):
            raise IOError('File doesnt exists: {}'.format(filepath))
        logger.info('Extracting video information...')
        # call to ffmpeg to get frame rate
        probe = Videos.get_info(filepath)
        fps = eval(probe['streams'][0]['avg_frame_rate'])
        n_frames = eval(probe['streams'][0]['nb_frames'])
        video_length = eval(probe['streams'][0]['duration'])
        logger.info('Video frame rate: {}[fps]'.format(fps))
        logger.info('Video number of frames: {}'.format(n_frames))
        logger.info('Video length in seconds: {}[s]'.format(video_length))

        # check split params and calc split params for ffmpeg
        if split_seconds is not None:
            # split by seconds
            split_length = split_seconds
            if split_length <= 0:
                raise ValueError('"split_length" can\'t be 0')
            split_count = int(np.ceil(video_length / split_length))
            list_frames_to_split = [fps * split_length * n for n in range(1, split_count)]
        elif split_chunks is not None:
            # split for known number of chunks
            split_count = split_chunks
            if split_chunks <= 0:
                raise ValueError('"split_chunks" size can\'t be 0')
            split_length = int(np.ceil(video_length / split_chunks))
            list_frames_to_split = [fps * split_length * n for n in range(1, split_count)]
        elif split_pairs is not None:
            if not isinstance(split_pairs, list):
                raise ValueError('"split_times" must be a list of tuples to split at.')
            if not (isinstance(split_pairs[0], list) or isinstance(split_pairs[0], tuple)):
                raise ValueError('"split_times" must be a list of tuples to split at.')
            list_frames_to_split = [fps * split_second for segment in split_pairs for split_second in segment]
            split_count = len(list_frames_to_split)
        else:
            raise ValueError('Must input one split option ("split_chunks", "split_time" or "split_pairs")')
        if split_count == 1:
            raise ValueError('Video length is less than the target split length.')
        # to integers
        list_frames_to_split = [int(i) for i in list_frames_to_split]
        # remove 0 if in the first segmetn
        if list_frames_to_split[0] == 0:
            list_frames_to_split.pop(0)
        # add last frames if not exists
        if list_frames_to_split[-1] != n_frames:
            list_frames_to_split = list_frames_to_split + [n_frames]
        logger.info('Splitting to %d chunks' % split_count)

        basename, ext = os.path.splitext(filepath)
        output_regex = os.path.join(basename, '%%03d.mp4')
        # create folder
        if not os.path.exists(basename):
            os.makedirs(basename, exist_ok=True)
        # run ffmpeg
        try:
            stream = ffmpeg.input(filepath, **{'loglevel': loglevel}).output(output_regex,
                                                                             **{'x264opts': 'bframes=0',
                                                                                'f': 'segment',
                                                                                'reset_timestamps': '1',
                                                                                'map': '0',
                                                                                'segment_frames': ','.join(
                                                                                    [str(i) for i in
                                                                                     list_frames_to_split])
                                                                                })
            ffmpeg.overwrite_output(stream).run(capture_stdout=True)
        except Exception:
            logger.exception('ffmpeg error in disassemble:')
            raise

        # split_cmd = 'ffmpeg -y -i "%s" -b 0 -f mp4 -reset_timestamps 1 -map 0 -f segment -segment_frames %s "%s"' % (
        #     filepath, ','.join([str(int(i)) for i in list_frames_to_split]), output_regex)
        # logger.info('About to run: %s' % split_cmd)
        # subprocess.check_call(shlex.split(split_cmd), universal_newlines=True)

        # rename
        list_frames_to_split = [0] + list_frames_to_split
        filenames = list()
        for n in range(split_count):
            old_filename = output_regex.replace('%03d', '%03d' % n)
            new_filename = output_regex.replace('%03d', '%s__%s' %
                                                (time.strftime('%H_%M_%S', time.gmtime(list_frames_to_split[n] / fps)),
                                                 time.strftime('%H_%M_%S',
                                                               time.gmtime(list_frames_to_split[n + 1] / fps))))
            filenames.append(new_filename)
            # rename to informative name
            if os.path.isfile(new_filename):
                logger.warning('File already exists. Overwriting!: {}'.format(new_filename))
                os.remove(new_filename)
            os.rename(old_filename, new_filename)
            # check if in pairs, if not - delete
            if split_pairs is not None:
                start_frames = [pair[0] for pair in split_pairs]
                end_frames = [pair[1] for pair in split_pairs]
                if (list_frames_to_split[n] // fps) in start_frames and (
                        list_frames_to_split[n + 1] // fps) in end_frames:
                    # keep video
                    pass
                else:
                    os.remove(new_filename)
        Videos.upload_to_platform(project_name=project_name,
                                  project_id=project_id,
                                  dataset_name=dataset_name,
                                  dataset_id=dataset_id,
                                  remote_path=remote_path,
                                  local_path=basename)

    @staticmethod
    def upload_to_platform(project_name=None, project_id=None, dataset_name=None, dataset_id=None,
                           local_path=None, remote_path=None):

        import dtlpy as dlp
        if project_id is not None or project_name is not None:
            project = dlp.projects.get(project_name=project_name, project_id=project_id)
            dataset = project.get(dataset_name=dataset_name, dataset_id=dataset_id)
            dataset.items.upload(dataset_name=dataset_name,
                                 dataset_id=dataset_id,
                                 local_path=local_path,
                                 remote_path=remote_path,
                                 file_types=['.mp4'])
        else:
            dataset = dlp.datasets.get(dataset_name=dataset_name, dataset_id=dataset_id)
            dataset.items.upload(local_path=local_path,
                                 remote_path=remote_path,
                                 file_types=['.mp4'])



def main(project_name, dataset_name):
    import traceback
    import random
    import string
    import dtlpy as dl
    from concurrent.futures import ThreadPoolExecutor

    def upload_single(w_url, w_metadata):
        try:
            item = dataset.items.upload(local_path=dl.UrlLink(
                ref=w_url,
                name='.{}.json'.format(
                    ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(10)))),
                remote_path='/cats',
                item_metadata=w_metadata)
            return item.id
        except:
            print(traceback.format_exc())
            return None

    primary_url = 'https://images.unsplash.com/photo-1518020382113-a7e8fc38eac9'
    secondary_urls = [
        'https://images.unsplash.com/photo-1543852786-1cf6624b9987',
        'https://images.unsplash.com/photo-1561948955-570b270e7c36'
    ]
    project = dl.projects.get(project_name=project_name)
    dataset = project.datasets.get(dataset_name=dataset_name)

    pool = ThreadPoolExecutor(max_workers=32)
    jobs = list()
    for i_url, url in enumerate(secondary_urls):
        jobs.append(pool.submit(upload_single, **{'w_url': url,
                                                  'w_metadata': {'user': {'num': i_url}}}))
    pool.shutdown()
    secondary_ids = [j.result() for j in jobs]
    modalities = list()
    for i_secondary_id, secondary_id in enumerate(secondary_ids):
        modalities.append(dl.Modality(modality_type=dl.ModalityTypeEnum.OVERLAY,
                                      ref=secondary_id,
                                      name='cat_num:{}'.format(i_secondary_id)).to_json())

    primary_item = dataset.items.upload(local_path=dl.UrlLink(ref=primary_url),
                                        item_metadata={'system': {'modalities': modalities}})



import collections.abc
import importlib.util
import inspect
import logging
import hashlib
import typing
import json
import sys
import os
import re
from copy import deepcopy
from shutil import copyfile
from concurrent.futures import ThreadPoolExecutor

from .. import entities, repositories, exceptions, utilities, miscellaneous, assets, services, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')

DEFAULT_PACKAGE_METHOD = entities.PackageFunction(name=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
                                                  description='',
                                                  inputs=[],
                                                  outputs=[])
DEFAULT_PACKAGE_MODULE = entities.PackageModule(init_inputs=list(),
                                                entry_point=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT,
                                                class_name=entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME,
                                                name=entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME,
                                                functions=[DEFAULT_PACKAGE_METHOD])


class PackageCatalog:
    DEFAULT_PACKAGE_TYPE = 'default_package_type'
    MULTI_MODULE = 'multi_module'
    MULTI_MODULE_WITH_TRIGGER = 'multi_module_with_trigger'
    SINGLE_FUNCTION_ITEM = 'single_function_item'
    SINGLE_FUNCTION_JSON = 'single_function_json'
    SINGLE_FUNCTION_DATASET = 'single_function_dataset'
    SINGLE_FUNCTION_ANNOTATION = 'single_function_annotation'
    SINGLE_FUNCTION_NO_INPUT = 'single_function_no_input'
    SINGLE_FUNCTION_MULTI_INPUT = 'single_function_multi_input'
    MULTI_FUNCTION_ITEM = 'multi_function_item'
    MULTI_FUNCTION_DATASET = 'multi_function_dataset'
    MULTI_FUNCTION_ANNOTATION = 'multi_function_annotation'
    MULTI_FUNCTION_NO_INPUT = 'multi_function_no_input'
    MULTI_FUNCTION_JSON = 'multi_function_json'
    SINGLE_FUNCTION_ITEM_WITH_TRIGGER = 'single_function_item_with_trigger'
    SINGLE_FUNCTION_DATASET_WITH_TRIGGER = 'single_function_dataset_with_trigger'
    SINGLE_FUNCTION_ANNOTATION_WITH_TRIGGER = 'single_function_annotation_with_trigger'
    MULTI_FUNCTION_ITEM_WITH_TRIGGERS = 'multi_function_item_with_triggers'
    MULTI_FUNCTION_DATASET_WITH_TRIGGERS = 'multi_function_dataset_with_triggers'
    MULTI_FUNCTION_ANNOTATION_WITH_TRIGGERS = 'multi_function_annotation_with_triggers'
    CONVERTER_FUNCTION = 'converter'


class Packages:
    """
    Packages Repository

    The Packages class allows users to manage packages (code used for running in Dataloop's FaaS) and their properties.
    Read more about `Packages <https://developers.dataloop.ai/tutorials/faas/introduction/chapter/>`_.
    """

    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project
        self.package_io = PackageIO()

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            try:
                self._project = repositories.Projects(client_api=self._client_api).get()
            except exceptions.NotFound:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Missing "project". need to set a Project entity or use project.packages repository')
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/packages".format(self.project.id))

    def open_in_web(self,
                    package: entities.Package = None,
                    package_id: str = None,
                    package_name: str = None):
        """
        Open the package in the web platform.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.package.Package package: package entity
        :param str package_id: package id
        :param str package_name: package name

        **Example**:

        .. code-block:: python

            project.packages.open_in_web(package_id='package_id')
        """
        if package_name is not None:
            package = self.get(package_name=package_name)
        if package is not None:
            package.open_in_web()
        elif package_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(package_id) + '/main')
        else:
            self._client_api._open_in_web(url=self.platform_url)

    @_api_reference.add(path='/packages/{id}/revisions', method='get')
    def revisions(self, package: entities.Package = None, package_id: str = None):
        """
        Get the package revision history.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.package.Package package: package entity
        :param str package_id: package id

        **Example**:

        .. code-block:: python

            project.packages.revisions(package='package_entity')
        """
        if package is None and package_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='must provide an identifier in inputs: "package" or "package_id"')
        if package is not None:
            package_id = package.id

        success, response = self._client_api.gen_request(
            req_type="get",
            path="/packages/{}/revisions".format(package_id))
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/packages/{id}', method='get')
    def get(self,
            package_name: str = None,
            package_id: str = None,
            checkout: bool = False,
            fetch=None,
            log_error=True) -> entities.Package:
        """
        Get Package object to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str package_id: package id
        :param str package_name: package name
        :param bool checkout: set the package as a default package object (cookies)
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :param bool log_error: log error if package not found
        :return: Package object
        :rtype: dtlpy.entities.package.Package

        **Example**:

        .. code-block:: python

            project.packages.get(package_id='package_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if package_name is None and package_id is None:
            package = self.__get_from_cache()
            if package is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out Package was found, must checkout or provide an identifier in inputs')
        elif fetch:
            if package_id is not None:
                success, response = self._client_api.gen_request(
                    req_type="get",
                    path="/packages/{}".format(package_id),
                    log_error=log_error)
                if not success:
                    raise exceptions.PlatformException(response)
                package = entities.Package.from_json(client_api=self._client_api,
                                                     _json=response.json(),
                                                     project=self._project)
                # verify input package name is same as the given id
                if package_name is not None and package.name != package_name:
                    logger.warning(
                        "Mismatch found in packages.get: package_name is different then package.name:"
                        " {!r} != {!r}".format(
                            package_name,
                            package.name))
            elif package_name is not None:
                filters = entities.Filters(field='name', values=package_name, resource=entities.FiltersResource.PACKAGE,
                                           use_defaults=False)
                if self._project is not None:
                    filters.add(field='projectId', values=self._project.id)
                packages = self.list(filters=filters)
                if packages.items_count == 0:
                    raise exceptions.PlatformException(
                        error='404',
                        message='Package not found. Name: {}'.format(package_name))
                elif packages.items_count > 1:
                    raise exceptions.PlatformException(
                        error='400',
                        message='More than one package found by the name of: {} '
                                'Please get package from a project entity'.format(package_name))
                package = packages.items[0]
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out Package was found, must checkout or provide an identifier in inputs')
        else:
            package = entities.Package.from_json(_json={'id': package_id,
                                                        'name': package_name},
                                                 client_api=self._client_api,
                                                 project=self._project,
                                                 is_fetched=False)

        if checkout:
            self.checkout(package=package)
        return package

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Package]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_package, package in enumerate(response_items):
            jobs[i_package] = pool.submit(entities.Package._protected_from_json,
                                          **{'client_api': self._client_api,
                                             '_json': package,
                                             'project': self._project})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        packages = miscellaneous.List([r[1] for r in results if r[0] is True])
        return packages

    def _list(self, filters: entities.Filters):
        url = '/query/faas'

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/query/faas', method='post')
    def list(self, filters: entities.Filters = None, project_id: str = None) -> entities.PagedEntities:
        """
        List project packages.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param str project_id: project id
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            project.packages.list()
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.PACKAGE)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.PACKAGE:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.PACKAGE. Got: {!r}'.format(filters.resource))

        if project_id is None and self._project is not None:
            project_id = self._project.id

        if project_id is not None:
            filters.add(field='projectId', values=project_id)

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       project_id=project_id,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def pull(self, package: entities.Package, version=None, local_path=None, project_id=None):
        """
        Pull (download) the package to a local path.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.package.Package package: package entity
        :param str version: the package version to pull
        :param local_path: the path of where to save the package
        :param project_id: the project id that include the package
        :return: local path where the package pull
        :rtype: str

        **Example**:

        .. code-block:: python

            project.packages.pull(package='package_entity', local_path='local_path')
        """
        if isinstance(version, int):
            logger.warning('Deprecation Warning - Package/service versions have been refactored'
                           'The version you provided has type: int, it will be converted to: 1.0.{}'
                           'Next time use a 3-level semver for package/service versions'.format(version))

        dir_version = version
        if version is None:
            dir_version = package.version

        if local_path is None:
            local_path = os.path.join(
                services.service_defaults.DATALOOP_PATH,
                "packages",
                "{}-{}".format(package.name, package.id),
                str(dir_version))

        if version is None or version == package.version:
            package_revision = package
        else:
            versions = [revision for revision in package.revisions if revision['version'] == version]
            if len(versions) <= 0:
                raise exceptions.PlatformException('404', 'Version not found: version={}'.format(version))
            elif len(versions) > 1:
                raise exceptions.PlatformException('404', 'More than one version found: version={}'.format(version))
            package_revision = entities.Package.from_json(
                _json=versions[0],
                project=package.project,
                client_api=package._client_api
            )
        if package_revision.codebase.type == entities.PackageCodebaseType.ITEM:
            local_path = package_revision.codebases.unpack(codebase_id=package_revision.codebase.item_id,
                                                           local_path=local_path)
        elif package_revision.codebase.type == entities.PackageCodebaseType.GIT:
            local_path = package_revision.codebases.unpack(codebase=package_revision.codebase,
                                                           local_path=local_path)
        else:
            raise Exception(
                'Pull can only be performed on packages with Item or Git codebase. Codebase type: {!r}'.format(
                    package_revision.codebase.type))

        return local_path

    def _name_validation(self, name: str):
        url = '/piper-misc/naming/packages/{}'.format(name)

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)

    @staticmethod
    def _validate_slots(slots, modules):
        for slot in slots:
            matched_module = [module for module in modules if slot.module_name == module.name]
            if len(matched_module) != 1:
                raise ValueError('Module {!r} in slots is not defined in modules.'.format(slot.module_name))
            matched_func = [func for func in matched_module[0].functions if slot.function_name == func.name]
            if len(matched_func) != 1:
                raise ValueError('Function {!r} in slots is not defined in module {!r}.'.format(slot.function_name,
                                                                                                slot.module_name))

    def build_requirements(self, filepath) -> list:
        """
        Build a requirement list (list of packages your code requires to run) from a file path. **The file listing the requirements MUST BE a txt file**.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param filepath: path of the requirements file
        :return: a list of dl.PackageRequirement
        :rtype: list
        """
        requirements_list = []
        if not os.path.exists(filepath):
            raise exceptions.PlatformException('400', 'requirements file {} does not exist'.format(filepath))
        with open(filepath) as requirements_txt:
            for requirement in requirements_txt:
                requirement = requirement.strip()
                if '\n' in requirement:
                    requirement.replace('\n', '')
                if requirement.startswith('#'):
                    continue
                elif ' #' in requirement:
                    requirement = requirement[:requirement.find(' #')]
                if 'install -r' in requirement:
                    requirements_list.extend(self.build_requirements(requirement[requirement.find('-r ') + 3:]))
                    continue
                requirement_spit = requirement.split(',')
                for req in requirement_spit:
                    name_version = re.split('<=|>=|==|<|>', req)
                    req_name = name_version[0]
                    req_version, op = None, None
                    if len(name_version) > 1:
                        op = re.findall('<=|>=|==|<|>', req)[0]
                        req_version = name_version[1]
                        if req_name == '':
                            req_name = last_req_name
                        last_req_name = req_name
                    requirements_list.append(
                        entities.PackageRequirement(name=req_name, version=req_version, operator=op))
        return requirements_list

    @_api_reference.add(path='/packages', method='post')
    def push(self,
             project: entities.Project = None,
             project_id: str = None,
             package_name: str = None,
             src_path: str = None,
             codebase: typing.Union[entities.GitCodebase, entities.ItemCodebase, entities.FilesystemCodebase] = None,
             modules: typing.List[entities.PackageModule] = None,
             is_global: bool = None,
             checkout: bool = False,
             revision_increment: str = None,
             version: str = None,
             ignore_sanity_check: bool = False,
             service_update: bool = False,
             service_config: dict = None,
             slots: typing.List[entities.PackageSlot] = None,
             requirements: typing.List[entities.PackageRequirement] = None,
             package_type=None,
             metadata=None
             ) -> entities.Package:
        """
        Push your local package to the UI.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        Project will be taken in the following hierarchy:
        project(input) -> project_id(input) -> self.project(context) -> checked out

        :param dtlpy.entities.project.Project project: optional - project entity to deploy to. default from context or checked-out
        :param str project_id: optional - project id to deploy to. default from context or checked-out
        :param str package_name: package name
        :param str src_path: path to package codebase
        :param dtlpy.entities.codebase.Codebase codebase: codebase object
        :param list modules: list of modules PackageModules of the package
        :param bool is_global: is package is global or local
        :param bool checkout: checkout package to local dir
        :param str revision_increment: optional - str - version bumping method - major/minor/patch - default = None
        :param str version: semver version f the package
        :param bool ignore_sanity_check: NOT RECOMMENDED - skip code sanity check before pushing
        :param bool service_update: optional - bool - update the service
        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :param list slots: optional - list of slots PackageSlot of the package
        :param list requirements: requirements - list of package requirements
        :param str package_type: default 'faas', options: 'app', 'ml
        :param dict metadata: dictionary of system and user metadata

        :return: Package object
        :rtype: dtlpy.entities.package.Package

        **Example**:

        .. code-block:: python

            project.packages.push(package_name='package_name',
                                    modules=[module],
                                    version='1.0.0',
                                    src_path=os.getcwd()
                                )
        """
        # get project
        project_to_deploy = None
        if project is not None:
            project_to_deploy = project
        elif project_id is not None:
            project_to_deploy = repositories.Projects(client_api=self._client_api).get(project_id=project_id)
        elif self._project is not None:
            project_to_deploy = self._project
        else:
            try:
                project_to_deploy = repositories.Projects(client_api=self._client_api).get()
            except Exception:
                pass

        if project_to_deploy is None:
            raise exceptions.PlatformException(
                error='400',
                message='Missing project from "packages.push" function. '
                        'Please provide project or id, use Packages from a '
                        'project.packages repository or checkout a project')

        # source path
        if src_path is None:
            if codebase is None:
                src_path = os.getcwd()
                logger.warning('No src_path is given, getting package information from cwd: {}'.format(src_path))
        else:
            # Validate src_path to prevent path traversal
            miscellaneous.PathUtils.validate_directory_path(
                src_path,
                base_path=os.getcwd(),
                must_exist=True
            )

        # get package json
        package_from_json = dict()
        path_files = os.listdir(src_path) if src_path is not None else []
        if assets.paths.PACKAGE_FILENAME in path_files:
            with open(os.path.join(src_path, assets.paths.PACKAGE_FILENAME), 'r') as f:
                package_from_json = json.load(f)

        if requirements is not None and not isinstance(requirements, list):
            requirements = [requirements]

        if requirements and assets.paths.REQUIREMENTS_FILENAME in path_files:
            logger.warning('Have both requirements param and requirements file will overwrite the requirements file')

        if not requirements and assets.paths.REQUIREMENTS_FILENAME in path_files:
            req_path = os.path.join(src_path, assets.paths.REQUIREMENTS_FILENAME)
            req_from_file = self.build_requirements(filepath=req_path)
            requirements = req_from_file

        # get name
        if package_name is None:
            package_name = package_from_json.get('name', 'default-package')

        if modules is None and 'modules' in package_from_json:
            modules = package_from_json['modules']

        if slots is None and 'slots' in package_from_json:
            slots = package_from_json['slots']

        if ignore_sanity_check:
            logger.warning(
                'Pushing a package without sanity check can cause errors when trying to deploy, '
                'trigger and execute functions.\n'
                'We highly recommend to not use the ignore_sanity_check flag')
        elif codebase is None:
            modules = self._sanity_before_push(src_path=src_path, modules=modules)

        self._name_validation(name=package_name)

        if slots is not None:
            if modules is None:
                raise ValueError('Cannot add slots when modules is empty.')
            if not isinstance(slots[0], entities.PackageSlot):
                slots = [entities.PackageSlot.from_json(_json=slot) for slot in slots]
            self._validate_slots(slots, modules)

        delete_codebase = False
        if codebase is None:
            codebase = project_to_deploy.codebases.pack(directory=src_path, name=package_name)
            delete_codebase = True

        try:
            # check if exist
            filters = entities.Filters(resource=entities.FiltersResource.PACKAGE, use_defaults=False)
            filters.add(field='projectId', values=project_to_deploy.id)
            filters.add(field='name', values=package_name)
            packages = self.list(filters=filters)
            if packages.items_count > 0:
                # package exists - need to update
                package = packages.items[0]

                if modules is not None:
                    package.modules = modules

                if slots is not None:
                    package.slots = slots

                if is_global is not None:
                    package.is_global = is_global

                if codebase is not None:
                    package.codebase = codebase

                if version is not None:
                    package.version = version

                if requirements is not None:
                    package.requirements = requirements

                if service_config is not None:
                    package.service_config = service_config

                if metadata is not None:
                    package.metadata = metadata

                if package_type is not None:
                    package.package_type = package_type
                package = self.update(package=package, revision_increment=revision_increment)
            else:
                package = self._create(
                    project_to_deploy=project_to_deploy,
                    package_name=package_name,
                    modules=modules,
                    slots=slots,
                    codebase=codebase,
                    is_global=is_global,
                    version=version,
                    service_config=service_config,
                    requirements=requirements,
                    package_type=package_type,
                    metadata=metadata
                )
            if checkout:
                self.checkout(package=package)
        except Exception:
            if delete_codebase:
                project_to_deploy.items.delete(item_id=codebase.item_id)
            raise

        logger.info("Package name: {!r}, id: {!r} has been added to project name: {!r}, id: {!r}".format(
            package.name,
            package.id,
            package.project.name,
            package.project.id))

        if service_update:
            service = package.services.get(service_name=package.name)
            service.package_revision = package.version
            service.update()
        return package

    def _create(self,
                project_to_deploy: entities.Project = None,
                codebase: typing.Union[entities.GitCodebase, entities.ItemCodebase, entities.FilesystemCodebase] = None,
                is_global: bool = None,
                package_name: str = entities.package_defaults.DEFAULT_PACKAGE_NAME,
                modules: typing.List[entities.PackageModule] = None,
                version: str = None,
                service_config: dict = None,
                slots: typing.List[entities.PackageSlot] = None,
                requirements: typing.List[entities.PackageRequirement] = None,
                package_type='faas',
                metadata=None
                ) -> entities.Package:
        """
        Create a package in platform.

        :param project_to_deploy:
        :param dtlpy.entities.codebase.Codebase codebase: codebase object
        :param bool is_global: is package is global or local
        :param str package_name: optional - default: 'default package'
        :param list modules: optional - PackageModules Entity
        :param str version: semver version of the package
        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :param list slots: optional - list of slots PackageSlot of the package
        :param list requirements: requirements - list of package requirements
        :param str package_type: default 'faas', options: 'app', 'ml
        :param dict metadata: dictionary of system and user metadata
        :return: Package object
        :rtype: dtlpy.entities.package.Package
        """
        if modules is not None:
            if not isinstance(modules, list):
                modules = [modules]

            if isinstance(modules[0], entities.PackageModule):
                modules = [module.to_json() for module in modules]

        if slots and isinstance(slots[0], entities.PackageSlot):
            slots = [slot.to_json() for slot in slots]

        if is_global is None:
            is_global = False

        payload = {'name': package_name,
                   'global': is_global,
                   'modules': modules,
                   'slots': slots,
                   'serviceConfig': service_config,
                   'type': package_type
                   }

        if codebase is not None:
            payload['codebase'] = codebase.to_json()

        if requirements is not None:
            payload['requirements'] = [r.to_json() for r in requirements]

        if version is not None:
            payload['version'] = version

        if metadata is not None:
            payload['metadata'] = metadata

        if project_to_deploy is not None:
            payload['projectId'] = project_to_deploy.id
        else:
            raise exceptions.PlatformException('400', 'Repository must have a project to perform this action')

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/packages',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Package.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=project_to_deploy)

    @_api_reference.add(path='/packages/{id}', method='delete')
    def delete(self, package: entities.Package = None, package_name=None, package_id=None):
        """
        Delete a Package object.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.package.Package package: package entity
        :param str package_id: package id
        :param str package_name: package name
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            project.packages.delete(package_name='package_name')
        """
        # get id and name
        if package_name is None or package_id is None:
            if package is None:
                package = self.get(package_id=package_id, package_name=package_name)
            package_id = package.id

        # check if project exist
        project_exists = True
        if self._project is None:
            try:
                if package is not None and package.project is not None:
                    self._project = package.project
                else:
                    self._project = repositories.Projects(client_api=self._client_api).get(
                        project_id=package.project_id)
            except exceptions.NotFound:
                project_exists = False

        if project_exists:
            # TODO can remove? in transactor?
            try:
                # create codebases repo
                codebases = repositories.Codebases(client_api=self._client_api, project=self._project)
                # get package codebases
                codebase_pages = codebases.list_versions(codebase_name=package_name)
                for codebase_page in codebase_pages:
                    for codebase in codebase_page:
                        codebase.delete()
            except exceptions.Forbidden:
                logger.debug('Failed to delete code-bases. Continue without')

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/packages/{}".format(package_id)
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return results
        return True

    @_api_reference.add(path='/packages/{id}', method='patch')
    def update(self, package: entities.Package, revision_increment: str = None) -> entities.Package:
        """
        Update Package changes to the platform.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.package.Package package:
        :param revision_increment: optional - str - version bumping method - major/minor/patch - default = None
        :return: Package object
        :rtype: dtlpy.entities.package.Package

        **Example**:

        .. code-block:: python

            project.packages.delete(package='package_entity')
        """

        if revision_increment is not None and isinstance(package.version, str) and len(package.version.split('.')) == 3:
            major, minor, patch = package.version.split('.')
            if revision_increment == 'patch':
                patch = int(patch) + 1
            elif revision_increment == 'minor':
                minor = int(minor) + 1
            elif revision_increment == 'major':
                major = int(major) + 1
            package.version = '{}.{}.{}'.format(major, minor, patch)

        # payload
        payload = package.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/packages/{}'.format(package.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Package.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=self._project)

    def deploy(self,
               package_id: str = None,
               package_name: str = None,
               package: entities.Package = None,
               service_name: str = None,
               project_id: str = None,
               revision: str or int = None,
               init_input: typing.Union[typing.List[entities.FunctionIO], entities.FunctionIO, dict] = None,
               runtime: typing.Union[entities.KubernetesRuntime, dict] = None,
               sdk_version: str = None,
               agent_versions: dict = None,
               bot: typing.Union[entities.Bot, str] = None,
               pod_type: entities.InstanceCatalog = None,
               verify: bool = True,
               checkout: bool = False,
               module_name: str = None,
               run_execution_as_process: bool = None,
               execution_timeout: int = None,
               drain_time: int = None,
               on_reset: str = None,
               max_attempts: int = None,
               force: bool = False,
               secrets: list = None,
               **kwargs) -> entities.Service:
        """
        Deploy a package. A service is required to run the code in your package.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str package_id: package id
        :param str package_name: package name
        :param dtlpy.entities.package.Package package: package entity
        :param str service_name: service name
        :param str project_id: project id
        :param str revision: package revision - default=latest
        :param init_input: config to run at startup
        :param dict runtime: runtime resources
        :param str sdk_version:  - optional - string - sdk version
        :param dict agent_versions: - dictionary - - optional -versions of sdk, agent runner and agent proxy
        :param str bot: bot email
        :param str pod_type: pod type dl.InstanceCatalog
        :param bool verify: verify the inputs
        :param bool checkout: checkout
        :param str module_name: module name
        :param bool run_execution_as_process: run execution as process
        :param int execution_timeout: execution timeout
        :param int drain_time: drain time
        :param str on_reset: on reset
        :param int max_attempts: Maximum execution retries in-case of a service reset
        :param bool force: optional - terminate old replicas immediately
        :param list secrets: list of the integrations ids
        :return: Service object
        :rtype: dtlpy.entities.service.Service

        **Example**:

        .. code-block:: python

            project.packages.deploy(service_name=package_name,
                                    execution_timeout=3 * 60 * 60,
                                    module_name=module.name,
                                    runtime=dl.KubernetesRuntime(
                                        concurrency=10,
                                        pod_type=dl.InstanceCatalog.REGULAR_S,
                                        autoscaler=dl.KubernetesRabbitmqAutoscaler(
                                            min_replicas=1,
                                            max_replicas=20,
                                            queue_length=20
                                        )
                                    )
                                )
        """

        if package is None:
            package = self.get(package_id=package_id, package_name=package_name)

        return package.services.deploy(package=package,
                                       service_name=service_name,
                                       revision=revision,
                                       init_input=init_input,
                                       runtime=runtime,
                                       sdk_version=sdk_version,
                                       agent_versions=agent_versions,
                                       project_id=project_id,
                                       pod_type=pod_type,
                                       bot=bot,
                                       verify=verify,
                                       module_name=module_name,
                                       checkout=checkout,
                                       jwt_forward=kwargs.get('jwt_forward', None),
                                       is_global=kwargs.get('is_global', None),
                                       run_execution_as_process=run_execution_as_process,
                                       execution_timeout=execution_timeout,
                                       drain_time=drain_time,
                                       on_reset=on_reset,
                                       max_attempts=max_attempts,
                                       force=force,
                                       secrets=secrets
                                       )

    def deploy_from_file(self, project, json_filepath):
        """
        Deploy package and service from a JSON file.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.project.Project project: project entity
        :param str json_filepath: path of the file to deploy
        :return: the package and the services

        **Example**:

        .. code-block:: python

            project.packages.deploy_from_file(project='project_entity', json_filepath='json_filepath')
        """
        # Validate file path to prevent path traversal
        miscellaneous.PathUtils.validate_file_path(json_filepath)
        with open(json_filepath, 'r') as f:
            data = json.load(f)

        package = project.packages.push(
            package_name=data['name'],
            modules=data['modules'],
            src_path=os.path.split(json_filepath)[0]
        )
        deployed_services = list()
        if 'services' in data:
            for service_json in data['services']:
                try:
                    # update
                    service_old = project.services.get(service_name=service_json['name'])
                except exceptions.NotFound:
                    # create
                    service_old = package.services.deploy(service_name=service_json['name'],
                                                          module_name=data['modules'][0]['name'])
                triggers = service_json.pop('triggers', None)
                artifacts = service_json.pop('artifacts', None)
                to_update, service = self.__compare_and_update_service_configurations(service=service_old,
                                                                                      service_json=service_json)
                if to_update:
                    service.package_revision = package.version
                    service.update()
                if triggers:
                    self.__compare_and_update_trigger_configurations(service, triggers)
                if artifacts:
                    self.__compare_and_upload_artifacts(artifacts, package)
                deployed_services.append(service)
        else:
            logger.warning(msg='Package JSON does not have services.')
        return deployed_services, package

    def __update_dict_recursive(self, d, u):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                inner_dict = d.get(k, {})
                if inner_dict is None:
                    inner_dict = {}
                d[k] = self.__update_dict_recursive(inner_dict, v)
            else:
                d[k] = v
        return d

    def __compare_and_update_trigger_configurations(self, service, trigger_list):
        """
        :param service:
        :param trigger_list:
        """
        service_triggers = service.triggers.list().all()
        triggers_dict = dict()
        for trigger in service_triggers:
            triggers_dict[trigger.name] = trigger

        for trigger_json in trigger_list:
            trigger_name = trigger_json['name']
            trigger_spec = trigger_json['spec']
            if trigger_name not in triggers_dict:
                # create
                service.triggers.create(
                    # general
                    name=trigger_name,
                    function_name=trigger_spec['operation']['functionName'],
                    trigger_type=trigger_json['type'],
                    # event
                    scope=trigger_json.get('scope', None),
                    is_global=trigger_json.get('global', None),
                    filters=trigger_spec.get('filter', None),
                    resource=trigger_spec.get('resource', None),
                    execution_mode=trigger_spec.get('executionMode', None),
                    actions=trigger_spec.get('actions', None),
                    # cron
                    start_at=trigger_spec.get('startAt', None),
                    end_at=trigger_spec.get('endAt', None),
                    cron=trigger_spec.get('cron', None),
                )
            else:
                existing_trigger = triggers_dict[trigger_name]
                # check diff
                _json = existing_trigger.to_json()
                # pop unmatched fields
                _ = _json['spec']['operation'].pop('serviceId')
                _new_json = _json.copy()
                # update fields from infra configurations
                self.__update_dict_recursive(_new_json, trigger_json)
                # check diffs
                diffs = miscellaneous.DictDiffer.diff(_json, _new_json)
                if diffs:
                    updated_trigger = existing_trigger.from_json(_json=_new_json,
                                                                 client_api=service._client_api,
                                                                 service=service,
                                                                 project=service._project)
                    updated_trigger.update()

    def __compare_and_update_service_configurations(self, service, service_json):
        """
        :param service:
        :param service_json:
        """
        # take json configuration from service
        _json = service.to_json()
        _new_json = _json.copy()
        # update fields from infra configurations
        self.__update_dict_recursive(_new_json, service_json)
        # check diffs
        diffs = miscellaneous.DictDiffer.diff(_json, _new_json)
        to_update = False
        if diffs:
            # build back service from json
            to_update = True
            service = service.from_json(_json=_new_json,
                                        client_api=service._client_api,
                                        package=service._package,
                                        project=service._project)
        return to_update, service

    def __compare_and_upload_artifacts(self, artifacts, package):
        """
        :param artifacts:
        :param package:
        """
        for artifact in artifacts:
            # create/get .dataloop dir
            cwd = os.getcwd()
            dl_dir = os.path.join(cwd, '.dataloop')
            if not os.path.isdir(dl_dir):
                os.mkdir(dl_dir)

            directory = os.path.abspath(artifact)
            m = hashlib.md5()
            with open(directory, 'rb') as f:
                for chunk in iter(lambda: f.read(4096), b''):
                    m.update(chunk)
            zip_md = m.hexdigest()
            artifacts_list = package.artifacts.list()
            for art in artifacts_list:
                if art.metadata['system']['md5'] == zip_md:
                    return

            artifact_item = package.artifacts.upload(filepath=directory,
                                                     package_name=package.name,
                                                     package=package,
                                                     overwrite=True, )

            artifact_item.metadata['system']['md5'] = zip_md
            artifact_item.update(True)

    @staticmethod
    def _package_json_generator(package_catalog, package_name):
        """
        :param package_catalog:
        :param package_name:
        """
        if package_catalog == PackageCatalog.DEFAULT_PACKAGE_TYPE:
            with open(assets.paths.ASSETS_PACKAGE_FILEPATH, 'r') as f:
                package_asset = json.load(f)
            package_asset['name'] = package_name
            return package_asset

        item_input = entities.FunctionIO(name='item', type='Item')
        annotation_input = entities.FunctionIO(name='annotation', type='Annotation')
        dataset_input = entities.FunctionIO(name='dataset', type='Dataset')
        json_input = entities.FunctionIO(name='config', type='Json')
        query_input = entities.FunctionIO(name='query', type='Json')

        func = entities.PackageFunction(name=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME)
        if package_catalog in [PackageCatalog.SINGLE_FUNCTION_ITEM, PackageCatalog.SINGLE_FUNCTION_ITEM_WITH_TRIGGER]:
            func.inputs = [item_input]
            modules = entities.PackageModule(functions=[func])
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_ANNOTATION,
                                 PackageCatalog.SINGLE_FUNCTION_ANNOTATION_WITH_TRIGGER]:
            func.inputs = [annotation_input]
            modules = entities.PackageModule(functions=[func])
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_DATASET,
                                 PackageCatalog.SINGLE_FUNCTION_DATASET_WITH_TRIGGER]:
            func.inputs = [dataset_input]
            modules = entities.PackageModule(functions=[func])
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_NO_INPUT]:
            modules = entities.PackageModule(functions=[func])
        elif package_catalog == PackageCatalog.SINGLE_FUNCTION_JSON:
            func.inputs = json_input
            modules = entities.PackageModule(functions=[func])
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_ITEM, PackageCatalog.MULTI_FUNCTION_ITEM_WITH_TRIGGERS]:
            func.inputs = [item_input]
            func.name = 'first_method'
            second_func = deepcopy(func)
            second_func.name = 'second_method'
            modules = entities.PackageModule(functions=[func, second_func])
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_ANNOTATION,
                                 PackageCatalog.MULTI_FUNCTION_ANNOTATION_WITH_TRIGGERS]:
            func.inputs = [annotation_input]
            func.name = 'first_method'
            second_func = deepcopy(func)
            second_func.name = 'second_method'
            modules = entities.PackageModule(functions=[func, second_func])
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_DATASET,
                                 PackageCatalog.MULTI_FUNCTION_DATASET_WITH_TRIGGERS]:
            func.inputs = [dataset_input]
            func.name = 'first_method'
            second_func = deepcopy(func)
            second_func.name = 'second_method'
            modules = entities.PackageModule(functions=[func, second_func])
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_JSON]:
            func.inputs = [json_input]
            func.name = 'first_method'
            second_func = deepcopy(func)
            second_func.name = 'second_method'
            modules = entities.PackageModule(functions=[func, second_func])
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_NO_INPUT]:
            func.name = 'first_method'
            second_func = deepcopy(func)
            second_func.name = 'second_method'
            modules = entities.PackageModule(functions=[func, second_func])
        elif package_catalog in [PackageCatalog.MULTI_MODULE, PackageCatalog.MULTI_MODULE_WITH_TRIGGER]:
            func.inputs = [item_input]
            module_a = entities.PackageModule(functions=[func],
                                              name='first_module',
                                              entry_point='first_module_class.py')
            module_b = entities.PackageModule(functions=[func],
                                              name='second_module',
                                              entry_point='second_module_class.py')
            modules = [module_a, module_b]
        elif package_catalog == PackageCatalog.SINGLE_FUNCTION_MULTI_INPUT:
            func.inputs = [item_input, dataset_input, json_input, annotation_input]
            modules = entities.PackageModule(functions=[func])
        elif package_catalog == PackageCatalog.CONVERTER_FUNCTION:
            func.inputs = [dataset_input, query_input]
            modules = entities.PackageModule(functions=[func])
        else:
            raise exceptions.PlatformException('404', 'Unknown package catalog type: {}'.format(package_catalog))

        if not isinstance(modules, list):
            modules = [modules]

        _json = {'name': package_name,
                 'modules': [module.to_json() for module in modules]}

        return _json

    @staticmethod
    def build_trigger_dict(actions,
                           name='default_module',
                           filters=None,
                           function='run',
                           execution_mode: entities.TriggerExecutionMode = 'Once',
                           type_t: entities.TriggerType = "Event"):
        """
        Build a trigger dictionary to trigger FaaS.
        Read more about `FaaS Triggers <https://developers.dataloop.ai/tutorials/faas/auto_annotate/chapter/#trigger-the-service>`_.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param actions: list of dl.TriggerAction
        :param str name: trigger name
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param str function: function name
        :param str execution_mode: execution mode dl.TriggerExecutionMode
        :param str type_t: trigger type dl.TriggerType
        :return: trigger dict
        :rtype: dict

        **Example**:

        .. code-block:: python

            project.packages.build_trigger_dict(actions=dl.TriggerAction.CREATED,
                                              function='run',
                                              execution_mode=dl.TriggerExecutionMode.ONCE)
        """
        if not isinstance(actions, list):
            actions = [actions]

        if not filters:
            filters = dict()

        trigger_dict = {
            "name": name,
            "type": type_t,
            "spec": {
                "filter": filters,
                "executionMode": execution_mode,
                "resource": "Item",
                "actions": actions,
                "input": {},
                "operation": {
                    "type": "function",
                    "functionName": function
                }
            }
        }
        return trigger_dict

    @staticmethod
    def _service_json_generator(package_catalog, service_name):
        """
        :param package_catalog:
        :param service_name:
        """
        triggers = list()
        with open(assets.paths.ASSETS_PACKAGE_FILEPATH, 'r') as f:
            service_json = json.load(f)["services"][0]
        if package_catalog == PackageCatalog.DEFAULT_PACKAGE_TYPE:
            triggers = service_json['triggers']
        elif 'triggers' in package_catalog:
            trigger_a = Packages.build_trigger_dict(name='first_trigger',
                                                    filters=dict(),
                                                    actions=['Created'],
                                                    function='first_method',
                                                    execution_mode=entities.TriggerExecutionMode.ONCE)
            trigger_b = Packages.build_trigger_dict(name='second_trigger',
                                                    filters=dict(),
                                                    actions=['Created'],
                                                    function='second_method',
                                                    execution_mode=entities.TriggerExecutionMode.ONCE)
            if 'item' in package_catalog:
                trigger_a['spec']['resource'] = trigger_b['spec']['resource'] = 'Item'
            if 'dataset' in package_catalog:
                trigger_a['spec']['resource'] = trigger_b['spec']['resource'] = 'Dataset'
            if 'annotation' in package_catalog:
                trigger_a['spec']['resource'] = trigger_b['spec']['resource'] = 'Annotation'
            triggers += [trigger_a, trigger_b]
        elif 'trigger' in package_catalog:
            trigger = Packages.build_trigger_dict(name='triggername',
                                                  filters=dict(),
                                                  actions=[],
                                                  function=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
                                                  execution_mode=entities.TriggerExecutionMode.ONCE)
            if 'item' in package_catalog:
                trigger['spec']['resource'] = 'Item'
            if 'dataset' in package_catalog:
                trigger['spec']['resource'] = 'Dataset'
            if 'annotation' in package_catalog:
                trigger['spec']['resource'] = 'Annotation'
            triggers += [trigger]

        if service_name is not None:
            service_json['name'] = service_name
        if 'multi_module' in package_catalog:
            service_json['moduleName'] = 'first_module'

        service_json['triggers'] = triggers

        return service_json

    @staticmethod
    def generate(name=None,
                 src_path: str = None,
                 service_name: str = None,
                 package_type=PackageCatalog.DEFAULT_PACKAGE_TYPE):
        """
        Generate a new package. Provide a file path to a JSON file with all the details of the package and service to generate the package. 

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str name: name
        :param str src_path: source file path
        :param str service_name: service name
        :param str package_type: package type from PackageCatalog

        **Example**:

        .. code-block:: python

            project.packages.generate(name='package_name',
                                      src_path='src_path')

        """
        # name
        if name is None:
            name = 'default-package'
        if service_name is None:
            service_name = 'default-service'

        # src path
        if src_path is None:
            src_path = os.getcwd()

        package_asset = Packages._package_json_generator(package_name=name,
                                                         package_catalog=package_type)
        services_asset = Packages._service_json_generator(package_catalog=package_type,
                                                          service_name=service_name)
        package_asset["services"] = [services_asset]
        to_generate = [
            {'src': assets.paths.ASSETS_GITIGNORE_FILEPATH, 'dst': os.path.join(src_path, '.gitignore'),
             'type': 'copy'},
            {'src': package_asset,
             'dst': os.path.join(src_path, assets.paths.PACKAGE_FILENAME), 'type': 'json'}
        ]

        if package_type == PackageCatalog.DEFAULT_PACKAGE_TYPE:
            to_generate.append({'src': assets.paths.ASSETS_MOCK_FILEPATH,
                                'dst': os.path.join(src_path, assets.paths.MOCK_FILENAME), 'type': 'copy'})
        else:
            module = entities.PackageModule.from_json(package_asset['modules'][0])
            function_name = module.functions[0].name

            to_generate.append({'src': Packages._mock_json_generator(module=module, function_name=function_name),
                                'dst': os.path.join(src_path, assets.paths.MOCK_FILENAME), 'type': 'json'})

        main_file_paths = Packages._entry_point_generator(package_catalog=package_type)
        if len(main_file_paths) == 1:
            to_generate.append({'src': main_file_paths[0],
                                'dst': os.path.join(src_path, assets.paths.MAIN_FILENAME),
                                'type': 'copy'})
        else:
            to_generate += [{'src': main_file_paths[0],
                             'dst': os.path.join(src_path, assets.paths.MODULE_A_FILENAME),
                             'type': 'copy'},
                            {'src': main_file_paths[1],
                             'dst': os.path.join(src_path, assets.paths.MODULE_B_FILENAME),
                             'type': 'copy'}
                            ]

        for job in to_generate:
            if not os.path.isfile(job['dst']):
                if job['type'] == 'copy':
                    copyfile(job['src'], job['dst'])
                elif job['type'] == 'json':
                    with open(job['dst'], 'w') as f:
                        json.dump(job['src'], f, indent=2)

        logger.info('Successfully generated package')

    def build(self, package: entities.Package, module_name=None, init_inputs=None, local_path=None, from_local=None):
        """
        Instantiate a module from the package code. Returns a loaded instance of the runner class

        :param package: Package entity
        :param module_name: Name of the module to build the runner class
        :param str init_inputs: dictionary of the class init variables (if exists). will be used to init the module class
        :param str local_path: local path of the package (if from_local=False - codebase will be downloaded)
        :param bool from_local: bool. if true - codebase will not be downloaded (only use local files)
        :return: dl.BaseServiceRunner
        """
        if module_name is None:
            module_name = entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME
        if init_inputs is None:
            init_inputs = dict()
        if package.codebase is None:
            raise exceptions.PlatformException(error='2001',
                                               message="Package {!r} has no codebase. unable to build a module".format(
                                                   package.name))

        if from_local is None:
            from_local = package.codebase.is_local

        if not from_local:
            # Not local => download codebase
            if package.codebase.is_local:
                raise RuntimeError("using local codebase: {}. Cannot use from_local=False".
                                   format(package.codebase.to_json()))
            local_path = self.pull(package=package, local_path=local_path)

        sys.path.append(local_path)  # TODO: is it the best way to use the imports?
        # load module from path
        module = [module for module in package.modules if module.name == module_name]
        if len(module) != 1:
            raise ValueError('Cannot find module from package. trying to load: {!r}, available modules: {}'.format(
                module_name,
                [module.name for module in package.modules]
            ))
        module = module[0]
        entry_point = os.path.join(local_path, module.entry_point)
        class_name = module.class_name
        if not os.path.isfile(entry_point):
            raise ValueError('Module entry point file is missing: {}'.format(entry_point))
        spec = importlib.util.spec_from_file_location(class_name, entry_point)
        if spec is None:
            raise ValueError('Cant load class ModelAdapter from entry point: {}'.format(entry_point))
        package_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(package_module)
        package_module_cls = getattr(package_module, class_name)
        package_module = package_module_cls(**init_inputs)
        logger.info('Model adapter loaded successfully!')
        return package_module

    @staticmethod
    def _mock_json_generator(module: entities.PackageModule, function_name):
        """
        :param module:
        :param function_name:
        """
        _json = dict(function_name=function_name, module_name=module.name)
        funcs = [func for func in module.functions if func.name == function_name]
        if len(funcs) == 1:
            func = funcs[0]
        else:
            raise exceptions.PlatformException('400', 'Other than 1 functions by the name of: {}'.format(function_name))
        _json['config'] = {inpt.name: entities.Package._mockify_input(input_type=inpt.type)
                           for inpt in module.init_inputs}
        _json['inputs'] = [{'name': inpt.name, 'value': entities.Package._mockify_input(input_type=inpt.type)}
                           for inpt in func.inputs]
        return _json

    @staticmethod
    def _entry_point_generator(package_catalog):
        """
        :param package_catalog:
        """
        if package_catalog == PackageCatalog.DEFAULT_PACKAGE_TYPE:
            paths_to_service_runner = assets.paths.ASSETS_MAIN_FILEPATH
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_ITEM, PackageCatalog.SINGLE_FUNCTION_ITEM_WITH_TRIGGER]:
            paths_to_service_runner = assets.service_runner_paths.SINGLE_METHOD_ITEM_SR_PATH
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_ANNOTATION,
                                 PackageCatalog.SINGLE_FUNCTION_ANNOTATION_WITH_TRIGGER]:
            paths_to_service_runner = assets.service_runner_paths.SINGLE_METHOD_ANNOTATION_SR_PATH
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_DATASET,
                                 PackageCatalog.SINGLE_FUNCTION_DATASET_WITH_TRIGGER]:
            paths_to_service_runner = assets.service_runner_paths.SINGLE_METHOD_DATASET_SR_PATH
        elif package_catalog in [PackageCatalog.SINGLE_FUNCTION_NO_INPUT]:
            paths_to_service_runner = assets.service_runner_paths.SINGLE_METHOD_SR_PATH
        elif package_catalog == PackageCatalog.SINGLE_FUNCTION_JSON:
            paths_to_service_runner = assets.service_runner_paths.SINGLE_METHOD_JSON_SR_PATH
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_ITEM, PackageCatalog.MULTI_FUNCTION_ITEM_WITH_TRIGGERS]:
            paths_to_service_runner = assets.service_runner_paths.MULTI_METHOD_ITEM_SR_PATH
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_ANNOTATION,
                                 PackageCatalog.MULTI_FUNCTION_ANNOTATION_WITH_TRIGGERS]:
            paths_to_service_runner = assets.service_runner_paths.MULTI_METHOD_ANNOTATION_SR_PATH
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_DATASET,
                                 PackageCatalog.MULTI_FUNCTION_DATASET_WITH_TRIGGERS]:
            paths_to_service_runner = assets.service_runner_paths.MULTI_METHOD_DATASET_SR_PATH
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_JSON]:
            paths_to_service_runner = assets.service_runner_paths.MULTI_METHOD_JSON_SR_PATH
        elif package_catalog in [PackageCatalog.MULTI_FUNCTION_NO_INPUT]:
            paths_to_service_runner = assets.service_runner_paths.MULTI_METHOD_SR_PATH
        elif package_catalog in [PackageCatalog.MULTI_MODULE, PackageCatalog.MULTI_MODULE_WITH_TRIGGER]:
            paths_to_service_runner = [assets.service_runner_paths.SINGLE_METHOD_ITEM_SR_PATH,
                                       assets.service_runner_paths.SINGLE_METHOD_ITEM_SR_PATH]
        elif package_catalog == PackageCatalog.SINGLE_FUNCTION_MULTI_INPUT:
            paths_to_service_runner = assets.service_runner_paths.SINGLE_METHOD_MULTI_INPUT_SR_PATH
        elif package_catalog == PackageCatalog.CONVERTER_FUNCTION:
            paths_to_service_runner = assets.service_runner_paths.CONVERTER_SR_PATH
        else:
            raise exceptions.PlatformException('404', 'Unknown package catalog type: {}'.format(package_catalog))

        if not isinstance(paths_to_service_runner, list):
            paths_to_service_runner = [paths_to_service_runner]

        return paths_to_service_runner

    @staticmethod
    def is_multithread(inputs):
        is_multi = False
        if len(inputs) > 0 and isinstance(inputs[0], list):
            is_multi = True

        if is_multi:
            for single_input in inputs:
                if not isinstance(single_input, list):
                    raise exceptions.PlatformException('400', 'mock.json inputs can be either list of dictionaries '
                                                              'or list of lists')

        return is_multi

    def test_local_package(self,
                           cwd: str = None,
                           concurrency: int = None,
                           package: entities.Package = None,
                           module_name: str = entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME,
                           function_name: str = entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
                           class_name: str = entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME,
                           entry_point: str = entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT,
                           mock_file_path: str = None):
        """
        Test local package in local environment.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str cwd: path to the file
        :param int concurrency: the concurrency of the test
        :param dtlpy.entities.package.Package package: entities.package
        :param str module_name: module name
        :param str function_name: function name
        :param str class_name: class name
        :param str entry_point: the file to run like main.py
        :param str mock_file_path: the mock file that have the inputs
        :return: list created by the function that tested the output
        :rtype: list

        **Example**:

        .. code-block:: python

            project.packages.test_local_package(cwd='path_to_package',
                                                package='package_entity',
                                                function_name='run')
        """
        if cwd is None:
            cwd = os.getcwd()
        if mock_file_path is None:
            mock_file_path = assets.paths.MOCK_FILENAME
        with open(os.path.join(cwd, mock_file_path), 'r') as f:
            mock_json = json.load(f)
        is_multithread = self.is_multithread(inputs=mock_json['inputs'])

        local_runner = LocalServiceRunner(self._client_api,
                                          packages=self,
                                          cwd=cwd,
                                          package=package,
                                          multithreading=is_multithread,
                                          concurrency=concurrency,
                                          module_name=module_name,
                                          function_name=function_name,
                                          class_name=class_name,
                                          entry_point=entry_point,
                                          mock_file_path=mock_file_path)

        if self._project is None:
            try:
                project = repositories.Projects(client_api=self._client_api).get()
            except Exception:
                project = None
        else:
            project = self.project

        return local_runner.run_local_project(project=project)

    def __get_from_cache(self) -> entities.Package:
        package = self._client_api.state_io.get('package')
        if package is not None:
            package = entities.Package.from_json(_json=package, client_api=self._client_api, project=self._project)
        return package

    def checkout(self,
                 package: entities.Package = None,
                 package_id: str = None,
                 package_name: str = None):
        """
        Checkout (switch) to a package.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.package.Package package: package entity
        :param str package_id: package id
        :param str package_name: package name

        **Example**:

        .. code-block:: python

            project.packages.checkout(package='package_entity')

        """
        if package is None:
            package = self.get(package_id=package_id, package_name=package_name)
        self._client_api.state_io.put('package', package.to_json())
        logger.info("Checked out to package {}".format(package.name))

    @staticmethod
    def check_cls_arguments(cls, missing, function_name, function_inputs):
        """
        Check class arguments. This method checks that the package function is correct.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param cls: packages class
        :param list missing: list of the missing params
        :param str function_name: name of function
        :param list function_inputs: list of function inputs
        """
        # input to function and inputs definitions match
        func = getattr(cls, function_name)
        func_inspect = inspect.getfullargspec(func)
        if not isinstance(function_inputs, list):
            function_inputs = [function_inputs]
        defined_input_names = [inp['name'] if isinstance(inp, dict) else inp.name for inp in function_inputs]

        args_with_values = dict()
        # go over the defaults and inputs from end to start to map between them
        if func_inspect.defaults is not None:
            for i_value in range(-1, -len(func_inspect.defaults) - 1, -1):
                args_with_values[func_inspect.args[i_value]] = func_inspect.defaults[i_value]

        for input_name in func_inspect.args:
            if input_name in ['self', 'progress', 'context']:
                continue
            if input_name not in defined_input_names:
                logger.warning('missing input name: "{}" in function definition: "{}"'.format(input_name,
                                                                                              function_name))
                if input_name not in args_with_values:
                    missing.append('missing input name: "{}" in function definition: "{}"'.format(input_name,
                                                                                                  function_name))

    @staticmethod
    def _sanity_before_push(src_path, modules):
        """
        :param src_path:
        :param modules:
        """
        if modules is None:
            modules = [DEFAULT_PACKAGE_MODULE]

        if not isinstance(modules, list):
            modules = [modules]

        if not isinstance(modules[0], entities.PackageModule):
            modules = [entities.PackageModule.from_json(_json=module) for module in modules]

        missing = list()
        for module in modules:
            entry_point = module.entry_point
            class_name = module.class_name

            check_init = True
            functions = module.functions
            # check in inputs is a list
            if not isinstance(functions, list):
                functions = [functions]

            for function in functions:
                # entry point exists
                entry_point_filepath = os.path.join(src_path, entry_point)
                if not os.path.isfile(entry_point_filepath):
                    missing.append('missing entry point file: {}'.format(entry_point_filepath))
                    continue

                # class name in entry point
                try:
                    spec = importlib.util.spec_from_file_location(class_name, entry_point_filepath)
                    cls_def = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(cls_def)
                except ImportError as e:
                    logger.warning(
                        'Cannot run sanity check to find conflicts between package module and source code'
                        'because some packages in your source code are not installed in your local environment'.format(
                            e.__str__())
                    )
                    return modules

                if not hasattr(cls_def, class_name):
                    missing.append('missing class: "{}" from file: "{}"'.format(class_name, entry_point_filepath))
                    continue

                # function in class
                cls = getattr(cls_def, class_name)
                if not hasattr(cls, function.name):
                    missing.append(
                        'missing function: "{}" from class: "{}"'.format(function.name, entry_point_filepath))
                    continue

                if check_init:
                    check_init = False
                    # input to __init__ and inputs definitions match
                    Packages.check_cls_arguments(cls=cls,
                                                 missing=missing,
                                                 function_name="__init__",
                                                 function_inputs=module.init_inputs)

                Packages.check_cls_arguments(cls=cls,
                                             missing=missing,
                                             function_name=function.name,
                                             function_inputs=function.inputs)

        if len(missing) != 0:
            raise ValueError('There are conflicts between modules and source code:'
                             '\n{}\n{}'.format(missing,
                                               'Please fix conflicts or use flag ignore_sanity_check'))

        return modules


class LocalServiceRunner:
    """
    Service Runner Class
    """

    def __init__(self,
                 client_api: ApiClient,
                 packages,
                 cwd=None,
                 multithreading=False,
                 concurrency=10,
                 package: entities.Package = None,
                 module_name=entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME,
                 function_name=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
                 class_name=entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME,
                 entry_point=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT,
                 mock_file_path=None):
        if cwd is None:
            self.cwd = os.getcwd()
        else:
            self.cwd = cwd

        self._client_api = client_api
        self._packages = packages
        self.package_io = PackageIO(cwd=self.cwd)
        self.multithreading = multithreading
        self.concurrency = concurrency
        self._module_name = module_name
        self._function_name = function_name
        self._entry_point = entry_point
        self._class_name = class_name
        self.package = package

        if mock_file_path is None:
            mock_file_path = 'mock.json'
        with open(os.path.join(self.cwd, mock_file_path), 'r') as f:
            self.mock_json = json.load(f)

    @property
    def function_name(self):
        return self.mock_json.get('function_name', self._function_name)

    @property
    def module_name(self):
        return self.mock_json.get('module_name', self._module_name)

    @property
    def class_name(self):
        return self.mock_json.get('class_name', self._class_name)

    @property
    def entry_point(self):
        return self.mock_json.get('entry_point', self._entry_point)

    def get_mainpy_run_service(self):
        """
        Get mainpy run service

        :return:
        """
        entry_point = os.path.join(self.cwd, self.entry_point)
        spec = importlib.util.spec_from_file_location(self.class_name, entry_point)
        cls_def = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(cls_def)
        service_runner = getattr(cls_def, self.class_name)

        if 'init_params' in self.mock_json:
            kwargs = self.mock_json.get('init_params')
        elif 'config' in self.mock_json:
            kwargs = self.mock_json.get('config')
        else:
            kwargs = dict()

        return service_runner(**kwargs)

    def run_local_project(self, project=None):
        """
        Run local project

        :param project: project entity
        """
        package_runner = self.get_mainpy_run_service()

        modules = self.package_io.get('modules') if self.package is None else [
            m.to_json() for m in self.package.modules
        ]

        if isinstance(modules, list) and len(modules) > 0:
            try:
                module = [module for module in modules if module['name'] == self.module_name][0]
            except Exception:
                raise exceptions.PlatformException(
                    '400', 'Module {} does not exist in package modules'.format(self.module_name)
                )
        else:
            module = DEFAULT_PACKAGE_MODULE.to_json()

        if isinstance(module['functions'], list) and len(module['functions']) > 0:
            try:
                func = [func for func in module['functions'] if func['name'] == self.function_name][0]
            except Exception:
                raise exceptions.PlatformException(
                    '400', 'function {} does not exist in package module'.format(self.function_name)
                )
        else:
            func = entities.PackageFunction(None, list(), list(), None).to_json()
        package_inputs = func['input']
        current_level = logging.root.level
        logging.root.setLevel('INFO')
        if not self.multithreading:
            kwargs = dict()
            progress = utilities.Progress()

            func = getattr(package_runner, self.function_name)
            params = list(inspect.signature(func).parameters)
            if "progress" in params:
                kwargs['progress'] = progress

            for package_input in package_inputs:
                kwargs[package_input['name']] = self.get_field(field_name=package_input['name'],
                                                               field_type=package_input['type'],
                                                               project=project, mock_json=self.mock_json)

            results = getattr(package_runner, self.function_name)(**kwargs)
        else:
            pool = ThreadPoolExecutor(max_workers=self.concurrency)
            inputs = self.mock_json['inputs']
            results = list()
            jobs = list()
            for single_input in inputs:
                kwargs = dict()
                progress = utilities.Progress()
                kwargs['progress'] = progress
                for package_input in package_inputs:
                    kwargs[package_input['name']] = self.get_field(field_name=package_input['name'],
                                                                   field_type=package_input['type'],
                                                                   project=project,
                                                                   mock_json=self.mock_json,
                                                                   mock_inputs=single_input)
                jobs.append(
                    pool.submit(
                        getattr(package_runner, self.function_name),
                        **kwargs
                    )
                )
            for job in jobs:
                results.append(job.result())

        logging.root.level = current_level
        return results

    @staticmethod
    def _is_entity(output_type: str):
        # output_type is a entity class starts with capital
        return hasattr(entities, output_type)

    def _fetch_single_resource(self, value, output_type):
        if isinstance(value, dict):
            params = {'{}_id'.format(output_type.lower()): value['{}_id'.format(output_type.lower())]}
        else:
            params = {'{}_id'.format(output_type.lower()): value}
        return getattr(repositories, '{}s'.format(output_type))(self._client_api).get(**params)

    def get_field(self, field_name, field_type, mock_json, project=None, mock_inputs=None):
        """
        Get field in mock json.

        :param field_name: field name
        :param field_type: field type
        :param mock_json: mock json
        :param project: project
        :param mock_inputs: mock inputs
        :return:
        """
        if mock_inputs is None:
            mock_inputs = mock_json['inputs']
        filtered_mock_inputs = list(filter(lambda input_field: input_field['name'] == field_name, mock_inputs))

        if len(filtered_mock_inputs) == 0:
            raise Exception('No entry for field {} found in mock'.format(field_name))
        if len(filtered_mock_inputs) > 1:
            raise Exception('Duplicate entries for field {} found in mock'.format(field_name))

        if field_type not in list(entities.PackageInputType):
            raise exceptions.PlatformException('400', 'Unknown resource type for field {}'.format(field_name))

        mock_input = filtered_mock_inputs[0]
        resource_id = mock_input['value']

        if field_type.endswith('[]') and self._is_entity(output_type=field_type.replace('[]', '')):
            field_type = field_type.replace('[]', '')
            value = resource_id if isinstance(resource_id, list) else [resource_id]
            return [
                self._fetch_single_resource(value=val, output_type=field_type) for val in value
            ]
        elif self._is_entity(output_type=field_type):
            return self._fetch_single_resource(value=resource_id, output_type=field_type)
        else:
            return resource_id


class PackageIO:

    def __init__(self, cwd=None):
        if cwd is None:
            cwd = os.getcwd()

        self.package_file_path = os.path.join(cwd, assets.paths.PACKAGE_FILENAME)

    def read_json(self):
        file_path = self.package_file_path

        with open(file_path, 'r') as fp:
            cfg = json.load(fp)
        return cfg

    def get(self, key):
        """
        :param key:
        """
        cfg = self.read_json()
        return cfg[key]

    def put(self, key, value):
        """
        :param key:
        :param value:
        """
        try:
            cfg = self.read_json()
            cfg[key] = value
            file_path = self.package_file_path
            with open(file_path, 'w') as fp:
                json.dump(cfg, fp, indent=2)
        except Exception:
            pass



def main():
    import dtlpy as dl
    
    project = dl.projects.get(project_name='Jungle')
    dataset = project.datasets.get(dataset_name='Tigers')

    converter = dl.Converter()
    converter.convert_dataset(dataset=dataset, to_format='yolo',
                              local_path='home/yolo_annotations/tigers')



from enum import Enum

import attr
import traceback
from collections import namedtuple

from .. import repositories, entities
from ..services.api_client import ApiClient


class FeatureEntityType(str, Enum):
    """Available types for Feature Set entities"""
    ITEM = 'item'
    ANNOTATION = 'annotation'
    DATASET = 'dataset'


@attr.s
class FeatureSet(entities.BaseEntity):
    """
    Com entity
    """
    # platform
    id = attr.ib()
    name = attr.ib()
    url = attr.ib(repr=False)
    creator = attr.ib(repr=False)
    created_at = attr.ib()
    updated_by = attr.ib()
    size = attr.ib()
    set_type = attr.ib()
    entity_type = attr.ib()
    project_id = attr.ib()
    model_id = attr.ib()
    org_id = attr.ib()

    # sdk
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['feature_sets', 'features'])
        feature_sets_repo = repositories.FeatureSets(client_api=self._client_api,
                                                     project_id=self.project_id)
        features_repo = repositories.Features(client_api=self._client_api,
                                              project_id=self.project_id,
                                              feature_set=self, )
        r = reps(feature_sets=feature_sets_repo,
                 features=features_repo)
        return r

    @property
    def feature_sets(self):
        assert isinstance(self._repositories.feature_sets, repositories.FeatureSets)
        return self._repositories.feature_sets

    @property
    def features(self):
        assert isinstance(self._repositories.features, repositories.Features)
        return self._repositories.features

    @staticmethod
    def _protected_from_json(_json, client_api, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: entity's object json
        :param client_api:
        :return:
        """
        try:
            feature_set = FeatureSet.from_json(_json=_json,
                                               client_api=client_api,
                                               is_fetched=is_fetched)
            status = True
        except Exception:
            feature_set = traceback.format_exc()
            status = False
        return status, feature_set

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Feature Set entity object from a json

        :param is_fetched: is Entity fetched from Platform
        :param _json: entity's object json
        :param client_api: client_api
        :return: Feature object
        """
        inst = cls(
            id=_json.get('id', None),
            name=_json.get('name', None),
            set_type=_json.get('type', None),
            entity_type=_json.get('entityType', None),
            size=_json.get('size', None),
            url=_json.get('url', None),
            project_id=_json.get('project', None),
            model_id=_json.get('modelId', None),
            created_at=_json.get('createdAt', None),
            creator=_json.get('creator', None),
            updated_by=_json.get('updatedBy', None),
            client_api=client_api,
            org_id=_json.get('org', None),
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """

        _json = {'id': self.id,
                 'type': self.set_type,
                 'entityType': self.entity_type,
                 'project': self.project_id,
                 'modelId': self.model_id,
                 'creator': self.creator,
                 'createdAt': self.created_at,
                 'updatedBy': self.updated_by,
                 'name': self.name,
                 'size': self.size,
                 'url': self.url}
        if self.org_id is not None:
            _json['org'] = self.org_id

        return _json

    def update(self):
        """
        Update feature set

        :return: entities.FeatureSet
        """
        return self.feature_sets.update(self)
    def delete(self):
        """
        Delete the feature set

        :return: success
        :rtype: bool
        """
        return self.feature_sets.delete(feature_set_id=self.id)
